#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""The module provides support for IEE for RTxxxx devices."""

import logging
import os
from copy import deepcopy
from struct import pack
from typing import Any, Dict, List, Optional, Union

from crcmod.predefined import mkPredefinedCrcFun

from spsdk import SPSDKError
from spsdk import version as spsdk_version
from spsdk.apps.utils.utils import filepath_from_config, get_key
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.crypto import (
    IEE_DATA_FOLDER,
    IEE_DATABASE_FILE,
    IEE_SCH_FILE,
    Counter,
    crypto_backend,
)
from spsdk.utils.database import Database
from spsdk.utils.easy_enum import Enum
from spsdk.utils.images import BinaryImage
from spsdk.utils.misc import (
    align_block,
    reverse_bytes_in_longs,
    split_data,
    value_to_bytes,
    value_to_int,
)
from spsdk.utils.registers import Registers
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas

logger = logging.getLogger(__name__)


class IeeKeyBlobLockAttributes(Enum):
    """IEE keyblob lock attributes."""

    LOCK = 0x95  #  IEE region lock.
    UNLOCK = 0x59  #  IEE region unlock.


class IeeKeyBlobKeyAttributes(Enum):
    """IEE keyblob key attributes."""

    CTR128XTS256 = 0x5A  # AES 128 bits (CTR), 256 bits (XTS)
    CTR256XTS512 = 0xA5  # AES 256 bits (CTR), 512 bits (XTS)


class IeeKeyBlobModeAttributes(Enum):
    """IEE Keyblob mode attributes."""

    Bypass = 0x6A  # AES encryption/decryption bypass
    AesXTS = 0xA6  # AES XTS mode
    AesCTRWAddress = 0x66  # AES CTR w address binding mode
    AesCTRWOAddress = 0xAA  # AES CTR w/o address binding mode
    AesCTRkeystream = 0x19  # AES CTR keystream only


class IeeKeyBlobWritePmsnAttributes(Enum):
    """IEE keblob write permission attributes."""

    ENABLE = 0x99  # Enable write permission in APC IEE
    DISABLE = 0x11  # Disable write permission in APC IEE


class IeeKeyBlobAttribute:
    """IEE Keyblob Attribute.

    typedef struct _iee_keyblob_attribute
    {
        uint8_t lock;      #  IEE Region Lock control flag.
        uint8_t keySize;   #  IEE AES key size.
        uint8_t aesMode;   #  IEE AES mode.
        uint8_t reserved;  #  Reserved.
    } iee_keyblob_attribute_t;
    """

    _FORMAT = "<BBBB"
    _SIZE = 4

    def __init__(
        self,
        lock: IeeKeyBlobLockAttributes,
        key_attribute: IeeKeyBlobKeyAttributes,
        aes_mode: IeeKeyBlobModeAttributes,
    ) -> None:
        """IEE keyblob constructor.

        :param lock: IeeKeyBlobLockAttributes
        :param key_attribute: IeeKeyBlobKeyAttributes
        :param aes_mode: IeeKeyBlobModeAttributes
        """
        self.lock = lock
        self.key_attribute = key_attribute
        self.aes_mode = aes_mode

    @property
    def ctr_mode(self) -> bool:
        """Return true if AES mode is CTR.

        :return: True if AES-CTR, false otherwise
        """
        if self.aes_mode in [
            IeeKeyBlobModeAttributes.AesCTRWAddress,
            IeeKeyBlobModeAttributes.AesCTRWOAddress,
            IeeKeyBlobModeAttributes.AesCTRkeystream,
        ]:
            return True
        return False

    @property
    def key1_size(self) -> int:
        """Return IEE key size based on selected mode.

        :return: Key size in bytes
        """
        if self.key_attribute == IeeKeyBlobKeyAttributes.CTR128XTS256:
            return 16
        return 32

    @property
    def key2_size(self) -> int:
        """Return IEE key size based on selected mode.

        :return: Key size in bytes
        """
        if self.key_attribute == IeeKeyBlobKeyAttributes.CTR128XTS256:
            return 16
        if self.ctr_mode:
            return 16
        return 32

    def export(self) -> bytes:
        """Export binary representation of KeyBlobAttribute.

        :rtype: serialized binary data
        """
        return pack(self._FORMAT, self.lock, self.key_attribute, self.aes_mode, 0)


class MasterId(Enum):
    """Master IDs for RT118x."""

    CM33 = 0b1000  # Cortex M33 Master ID
    CM7 = 0b1001  # Cortex M7 Master ID


class IeeKeyBlob:
    """IEE KeyBlob.

    typedef struct _iee_keyblob_
    {
        uint32_t header;                   #  IEE Key Blob header tag.
        uint32_t version;                  #  IEE Key Blob version, upward compatible.
        iee_keyblob_attribute_t attribute; #  IEE configuration attribute.
        uint32_t pageOffset;               #  IEE page offset.
        uint32_t key1[IEE_MAX_AES_KEY_SIZE_IN_BYTE /
                      sizeof(uint32_t)]; #  Encryption key1 for XTS-AES mode, encryption key for AES-CTR mode.
        uint32_t key2[IEE_MAX_AES_KEY_SIZE_IN_BYTE /
                      sizeof(uint32_t)]; #  Encryption key2 for XTS-AES mode, initial counter for AES-CTR mode.
        uint32_t startAddr;              #  Physical address of encryption region.
        uint32_t endAddr;                #  Physical address of encryption region.
        uint32_t reserved;               #  Reserved word.
        uint32_t crc32;                  #  Entire IEE Key Blob CRC32 value. Must be the last struct member.
    } iee_keyblob_t
    """

    _FORMAT = "LL4BL8L8LLLLL96B"

    HEADER_TAG = 0x49454542
    # Tag used in keyblob header
    # (('I' << 24) | ('E' << 16) | ('E' << 8) | ('B' << 0))
    KEYBLOB_VERSION = 0x56010000
    # Identifier of IEE keyblob version
    # (('V' << 24) | (1 << 16) | (0 << 8) | (0 << 0))
    KEYBLOB_OFFSET = 0x1000

    _IEE_ENCR_BLOCK_SIZE_XTS = 0x1000

    _ENCRYPTION_BLOCK_SIZE = 0x10

    _START_ADDR_MASK = 0x400 - 1
    # Region addresses are modulo 1024

    _END_ADDR_MASK = 0x3F8

    def __init__(
        self,
        attributes: IeeKeyBlobAttribute,
        start_addr: int,
        end_addr: int,
        key1: Optional[bytes] = None,
        key2: Optional[bytes] = None,
        page_offset: int = 0,
        crc: Optional[bytes] = None,
        master: Optional[MasterId] = None,
    ):
        """Constructor.

        :param attributes: IEE keyblob attributes
        :param start_addr: start address of the region
        :param end_addr: end address of the region
        :param key1: Encryption key1 for XTS-AES mode, encryption key for AES-CTR mode.
        :param key2: Encryption key2 for XTS-AES mode, initial_counter for AES-CTR mode.
        :param crc: optional value for unused CRC fill (for testing only); None to use calculated value
        :raises SPSDKError: Start or end address are not aligned
        :raises SPSDKError: When there is invalid key
        :raises SPSDKError: When there is invalid start/end address
        """
        self.attributes = attributes

        if key1 is None:
            key1 = crypto_backend().random_bytes(self.attributes.key1_size)
        if key2 is None:
            key2 = crypto_backend().random_bytes(self.attributes.key2_size)

        key1 = value_to_bytes(key1, byte_cnt=self.attributes.key1_size)
        key2 = value_to_bytes(key2, byte_cnt=self.attributes.key2_size)

        if master == MasterId.CM33:
            start_addr &= ~(1 << 28)
            end_addr &= ~(1 << 28)

        if start_addr < 0 or start_addr > end_addr or end_addr > 0xFFFFFFFF:
            raise SPSDKError("Invalid start/end address")

        if (start_addr & self._START_ADDR_MASK) != 0:
            raise SPSDKError(
                f"Start address must be aligned to {hex(self._START_ADDR_MASK + 1)} boundary"
            )

        self.master = master
        self.start_addr = start_addr
        self.end_addr = end_addr

        self.key1 = key1
        self.key2 = key2
        self.page_offset = page_offset

        self.crc_fill = crc

    def info(self) -> str:
        """Text info about the instance."""
        msg = ""
        msg += f"KEY 1:        {self.key1.hex()}\n"
        msg += f"KEY 2:       {self.key2.hex()}\n"
        msg += f"Start Addr: {hex(self.start_addr)}\n"
        msg += f"End Addr:   {hex(self.end_addr)}\n"
        return msg

    def plain_data(self) -> bytes:
        """Plain data for selected key range.

        :return: key blob exported into binary form (serialization)
        """
        result = bytes()
        result += pack("<II", self.HEADER_TAG, self.KEYBLOB_VERSION)
        result += self.attributes.export()
        result += pack("<I", self.page_offset)
        result += align_block(self.key1, 32)
        result += align_block(self.key2, 32)
        result += pack("<III", self.start_addr, self.end_addr, 0)
        crc: bytes = mkPredefinedCrcFun("crc-32-mpeg")(result).to_bytes(4, "little")
        result += crc

        return result

    def contains_addr(self, addr: int) -> bool:
        """Whether key blob contains specified address.

        :param addr: to be tested
        :return: True if yes, False otherwise
        """
        return self.start_addr <= addr <= self.end_addr

    def matches_range(self, image_start: int, image_end: int) -> bool:
        """Whether key blob matches address range of the image to be encrypted.

        :param image_start: start address of the image
        :param image_end: last address of the image
        :return: True if yes, False otherwise
        """
        return self.contains_addr(image_start) and self.contains_addr(image_end)

    def encrypt_image_xts(self, base_address: int, data: bytes) -> bytes:
        """Encrypt specified data using AES-XTS.

        :param base_address: of the data in target memory; must be >= self.start_addr
        :param data: to be encrypted (e.g. plain image); base_address + len(data) must be <= self.end_addr
        :return: encrypted data
        """
        encrypted_data = bytes()
        current_start = base_address
        key1 = reverse_bytes_in_longs(self.key1)
        key2 = reverse_bytes_in_longs(self.key2)

        for block in split_data(bytearray(data), self._IEE_ENCR_BLOCK_SIZE_XTS):
            tweak = self.calculate_tweak(current_start, self.master)

            encrypted_block = crypto_backend().aes_xts_encrypt(
                key1 + key2,
                block,
                tweak,
            )
            encrypted_data += encrypted_block
            current_start += len(block)

        return encrypted_data

    def encrypt_image_ctr(self, base_address: int, data: bytes) -> bytes:
        """Encrypt specified data using AES-CTR.

        :param base_address: of the data in target memory; must be >= self.start_addr
        :param data: to be encrypted (e.g. plain image); base_address + len(data) must be <= self.end_addr
        :return: encrypted data
        """
        encrypted_data = bytes()
        key = reverse_bytes_in_longs(self.key1)
        nonce = reverse_bytes_in_longs(self.key2)

        if self.master:
            base_address |= self.master << 32

        counter = Counter(nonce, ctr_value=base_address >> 4, ctr_byteorder_encoding="big")

        for block in split_data(bytearray(data), self._ENCRYPTION_BLOCK_SIZE):
            encrypted_block = crypto_backend().aes_ctr_encrypt(
                key,
                block,
                counter.value,
            )
            encrypted_data += encrypted_block
            counter.increment(self._ENCRYPTION_BLOCK_SIZE >> 4)

        return encrypted_data

    def encrypt_image(self, base_address: int, data: bytes) -> bytes:
        """Encrypt specified data.

        :param base_address: of the data in target memory; must be >= self.start_addr
        :param data: to be encrypted (e.g. plain image); base_address + len(data) must be <= self.end_addr
        :return: encrypted data
        :raises SPSDKError: If start address is not valid
        :raises NotImplementedError: AES-CTR is not implemented yet
        """
        if base_address % 16 != 0:
            raise SPSDKError("Invalid start address")  # Start address has to be 16 byte aligned
        data = align_block(data, self._ENCRYPTION_BLOCK_SIZE)  # align data length
        data_len = len(data)

        # check start and end addresses
        if not self.matches_range(base_address, base_address + data_len - 1):
            logger.warning(
                f"Image address range is not within key blob: {hex(self.start_addr)}-{hex(self.end_addr)}."
            )

        if self.attributes.ctr_mode:
            return self.encrypt_image_ctr(base_address, data)
        return self.encrypt_image_xts(base_address, data)

    @staticmethod
    def calculate_tweak(address: int, master: Optional[MasterId] = None) -> bytes:
        """Calculate tweak value for AES-XTS encryption based on the address value.

        :param address: start address of encryption
        :param master: MasterID
        :return: 16 byte tweak values
        """
        if master:
            address |= master << 32

        sector = address >> 12
        tweak = bytearray(16)
        for n in range(16):
            tweak[n] = sector & 0xFF
            sector = sector >> 8
        return bytes(tweak)


class Iee:
    """IEE: Inline Encryption Engine."""

    IEE_DATA_UNIT = 0x1000
    IEE_KEY_BLOBS_SIZE = 384

    def __init__(self) -> None:
        """Constructor."""
        self._key_blobs: List[IeeKeyBlob] = []

    def __getitem__(self, index: int) -> IeeKeyBlob:
        return self._key_blobs[index]

    def __setitem__(self, index: int, value: IeeKeyBlob) -> None:
        self._key_blobs.remove(self._key_blobs[index])
        self._key_blobs.insert(index, value)

    def add_key_blob(self, key_blob: IeeKeyBlob) -> None:
        """Add key for specified address range.

        :param key_blob: to be added
        """
        self._key_blobs.append(key_blob)

    def encrypt_image(self, image: bytes, base_addr: int) -> bytes:
        """Encrypt image with all available keyblobs.

        :param image: plain image to be encrypted
        :param base_addr: where the image will be located in target processor
        :return: encrypted image
        """
        encrypted_data = bytearray(image)
        addr = base_addr
        for block in split_data(image, self.IEE_DATA_UNIT):
            for key_blob in self._key_blobs:
                if key_blob.matches_range(addr, addr + len(block)):
                    logger.debug(
                        f"Encrypting {hex(addr)}:{hex(len(block) + addr)}"
                        f" with keyblob: \n {key_blob.info()}"
                    )
                    encrypted_data[
                        addr - base_addr : len(block) + addr - base_addr
                    ] = key_blob.encrypt_image(addr, block)
            addr += len(block)

        return bytes(encrypted_data)

    def get_key_blobs(self) -> bytes:
        """Get key blobs.

        :return: Binary key blobs joined together
        """
        result = bytes()
        for key_blob in self._key_blobs:
            result += key_blob.plain_data()

        # return result
        return align_block(result, self.IEE_KEY_BLOBS_SIZE)

    def encrypt_key_blobs(
        self,
        ibkek1: Union[bytes, str],
        ibkek2: Union[bytes, str],
        keyblob_address: int,
    ) -> bytes:
        """Encrypt keyblobs and export them as binary.

        :param ibkek1: key encryption key AES-XTS 256 bit
        :param ibkek2: key encryption key AES-XTS 256 bit
        :param keyblob_address: keyblob base address
        :return: encrypted keyblobs
        """
        plain_key_blobs = self.get_key_blobs()

        ibkek1 = reverse_bytes_in_longs(value_to_bytes(ibkek1, byte_cnt=32))
        logger.debug(f"IBKEK1: {' '.join(f'{b:02x}' for b in ibkek1)}")
        ibkek2 = reverse_bytes_in_longs(value_to_bytes(ibkek2, byte_cnt=32))
        logger.debug(f"IBKEK2 {' '.join(f'{b:02x}' for b in ibkek2)}")

        tweak = IeeKeyBlob.calculate_tweak(keyblob_address)
        return crypto_backend().aes_xts_encrypt(
            ibkek1 + ibkek2,
            plain_key_blobs,
            tweak,
        )


class IeeNxp(Iee):
    """IEE: Inline Encryption Engine."""

    def __init__(
        self,
        family: str,
        keyblob_address: int,
        ibkek1: Union[bytes, str],
        ibkek2: Union[bytes, str],
        key_blobs: Optional[List[IeeKeyBlob]] = None,
        binaries: Optional[BinaryImage] = None,
    ) -> None:
        """Constructor.

        :param family: Device family
        :param ibkek1: 256 bit key to encrypt IEE keyblob
        :param ibkek2: 256 bit key to encrypt IEE keyblob
        :param key_blobs: Optional Key blobs to add to IEE, defaults to None
        :raises SPSDKValueError: Unsupported family
        """
        super().__init__()

        if family not in self.get_supported_families():
            raise SPSDKValueError(f"Unsupported family{family} by IEE")

        self.family = family
        self.ibkek1 = bytes.fromhex(ibkek1) if isinstance(ibkek1, str) else ibkek1
        self.ibkek2 = bytes.fromhex(ibkek2) if isinstance(ibkek2, str) else ibkek2
        self.keyblob_address = keyblob_address
        self.binaries = binaries

        self.database = Database(IEE_DATABASE_FILE)
        self.blobs_min_cnt = self.database.get_device_value("key_blob_min_cnt", device=family)
        self.blobs_max_cnt = self.database.get_device_value("key_blob_max_cnt", device=family)
        self.generate_keyblob = self.database.get_device_value("generate_keyblob", device=family)

        if key_blobs:
            for key_blob in key_blobs:
                self.add_key_blob(key_blob)

    def export_key_blobs(self) -> bytes:
        """Export encrypted keyblobs in binary.

        :return: Encrypted keyblobs
        """
        return self.encrypt_key_blobs(self.ibkek1, self.ibkek2, self.keyblob_address)

    def export_image(self) -> Optional[BinaryImage]:
        """Export encrypted image.

        :return: Encrypted image
        """
        if self.binaries is None:
            return None
        self.binaries.validate()

        binaries: BinaryImage = deepcopy(self.binaries)

        for binary in binaries.sub_images:
            if binary.binary:
                binary.binary = self.encrypt_image(
                    binary.binary, binary.absolute_address + self.keyblob_address
                )
            for segment in binary.sub_images:
                if segment.binary:
                    segment.binary = self.encrypt_image(
                        segment.binary,
                        segment.absolute_address + self.keyblob_address,
                    )

        binaries.validate()
        return binaries

    def get_blhost_script_otp_kek(self) -> str:
        """Create BLHOST script to load fuses needed to run IEE with OTP fuses.

        :return: BLHOST script that loads the keys into fuses.
        """
        if not self.database.get_device_value("has_kek_fuses", self.family, default=False):
            logger.debug(f"The {self.family} has no IEE KEK fuses")
            return ""

        xml_fuses = self.database.get_device_value("reg_fuses", device=self.family, default=None)
        if not xml_fuses:
            logger.debug(f"The {self.family} has no IEE fuses definition")
            return ""

        xml_fuses = os.path.join(IEE_DATA_FOLDER, xml_fuses)

        fuses = Registers(self.family, base_endianness="little")
        grouped_regs = self.database.get_device_value(
            "grouped_registers", device=self.family, default=None
        )

        fuses.load_registers_from_xml(xml_fuses, grouped_regs=grouped_regs)
        fuses.find_reg("USER_KEY1").set_value(self.ibkek1)
        fuses.find_reg("USER_KEY2").set_value(self.ibkek2)

        load_iee = fuses.find_reg("LOAD_IEE_KEY")
        load_iee.find_bitfield("LOAD_IEE_KEY_BITFIELD").set_value(1)

        encrypt_engine = fuses.find_reg("ENCRYPT_XIP_ENGINE")
        encrypt_engine.find_bitfield("ENCRYPT_XIP_ENGINE_BITFIELD").set_value(1)

        boot_cfg = fuses.find_reg("BOOT_CFG")
        boot_cfg.find_bitfield("ENCRYPT_XIP_EN_BITFIELD").set_value(1)

        ibkek_lock = fuses.find_reg("USER_KEY_RLOCK")
        ibkek_lock.find_bitfield("USER_KEY1_RLOCK").set_value(1)
        ibkek_lock.find_bitfield("USER_KEY2_RLOCK").set_value(1)

        ret = (
            "# BLHOST IEE fuses programming script\n"
            f"# Generated by SPSDK {spsdk_version}\n"
            f"# Chip: {self.family} \n\n"
        )

        ret += f"# OTP IBKEK1: {self.ibkek1.hex()}\n\n"
        for reg in fuses.find_reg("USER_KEY1").sub_regs:
            ret += f"# {reg.name} fuse.\n"
            ret += f"efuse-program-once {hex(reg.offset)} 0x{reg.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n\n# OTP IBKEK2: {self.ibkek2.hex()}\n\n"
        for reg in fuses.find_reg("USER_KEY2").sub_regs:
            ret += f"# {reg.name} fuse.\n"
            ret += f"efuse-program-once {hex(reg.offset)} 0x{reg.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n\n# {load_iee.name} fuse.\n"
        for bitfield in load_iee.get_bitfields():
            ret += f"#   {bitfield.name}: {bitfield.get_enum_value()}\n"
        ret += f"efuse-program-once {hex(load_iee.offset)} 0x{load_iee.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n\n# {encrypt_engine.name} fuse.\n"
        for bitfield in encrypt_engine.get_bitfields():
            ret += f"#   {bitfield.name}: {bitfield.get_enum_value()}\n"
        ret += (
            f"efuse-program-once {hex(encrypt_engine.offset)} "
            f"0x{encrypt_engine.get_bytes_value().hex()} --no-verify\n"
        )

        ret += f"\n\n# {ibkek_lock.name} fuse.\n"
        for bitfield in ibkek_lock.get_bitfields():
            ret += f"#   {bitfield.name}: {bitfield.get_enum_value()}\n"
        ret += f"efuse-program-once {hex(ibkek_lock.offset)} 0x{ibkek_lock.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n\n# {boot_cfg.name} fuse.\n"
        ret += "WARNING!! Check SRM and set all desired bitfields for boot configuration"
        for bitfield in boot_cfg.get_bitfields():
            ret += f"#   {bitfield.name}: {bitfield.get_enum_value()}\n"
        ret += (
            f"# efuse-program-once {hex(boot_cfg.offset)} "
            f"0x{boot_cfg.get_bytes_value().hex()} --no-verify\n"
        )

        return ret

    def binary_image(
        self,
        plain_data: bool = False,
        data_alignment: int = 16,
        keyblob_name: str = "iee_keyblob.bin",
        image_name: str = "encrypted.bin",
    ) -> BinaryImage:
        """Get the IEE Binary Image representation.

        :param plain_data: Binary representation in plain format, defaults to False
        :param data_alignment: Alignment of data part key blobs.
        :param keyblob_name: Filename of the IEE keyblob
        :param image_name: Filename of the IEE image
        :return: IEE in BinaryImage.
        """
        iee = BinaryImage(image_name, offset=self.keyblob_address)
        if self.generate_keyblob:
            # Add mandatory IEE keyblob
            iee_keyblobs = self.get_key_blobs() if plain_data else self.export_key_blobs()
            iee.add_image(
                BinaryImage(
                    keyblob_name,
                    offset=0,
                    description=f"IEE keyblobs {self.family}",
                    binary=iee_keyblobs,
                )
            )
        binaries = self.export_image()

        if binaries:
            binaries.alignment = data_alignment
            binaries.validate()
            iee.add_image(binaries)

        return iee

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get all supported families for AHAB container.

        :return: List of supported families.
        """
        database = Database(IEE_DATABASE_FILE)
        return database.devices.device_names

    @staticmethod
    def get_validation_schemas(family: str) -> List[Dict[str, Any]]:
        """Get list of validation schemas.

        :param family: Family for which the template should be generated.
        :return: Validation list of schemas.
        """
        if family not in IeeNxp.get_supported_families():
            return []

        database = Database(IEE_DATABASE_FILE)
        schemas = ValidationSchemas.get_schema_file(IEE_SCH_FILE)
        family_sch = deepcopy(schemas["iee_family"])
        family_sch["properties"]["family"]["enum"] = IeeNxp.get_supported_families()
        family_sch["properties"]["family"]["template_value"] = family
        ret = [family_sch, schemas["iee_output"], schemas["iee"]]
        additional_schemes = database.get_device_value("additional_template", family, default=[])
        ret.extend([schemas[x] for x in additional_schemes])
        return ret

    @staticmethod
    def get_validation_schemas_family() -> List[Dict[str, Any]]:
        """Get list of validation schemas for family key.

        :return: Validation list of schemas.
        """
        schemas = ValidationSchemas.get_schema_file(IEE_SCH_FILE)
        family_sch = deepcopy(schemas["iee_family"])
        family_sch["properties"]["family"]["enum"] = IeeNxp.get_supported_families()
        return [family_sch]

    @staticmethod
    def generate_config_template(family: str) -> Dict[str, Any]:
        """Generate IEE configuration template.

        :param family: Family for which the template should be generated.
        :return: Dictionary of individual templates (key is name of template, value is template itself).
        """
        val_schemas = IeeNxp.get_validation_schemas(family)
        database = Database(IEE_DATABASE_FILE)

        if val_schemas:
            template_note = database.get_device_value(
                "additional_template_text", family, default=""
            )
            title = f"IEE: Inline Encryption Engine Configuration template for {family}."

            yaml_data = ConfigTemplate(title, val_schemas, note=template_note).export_to_yaml()

            return {f"{family}_iee": yaml_data}

        return {}

    @staticmethod
    def load_from_config(
        config: Dict[str, Any], config_dir: str, search_paths: Optional[List[str]] = None
    ) -> "IeeNxp":
        """Converts the configuration option into an IEE image object.

        "config" content array of containers configurations.

        :param config: array of IEE configuration dictionaries.
        :param config_dir: directory where the config is located
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: initialized IEE object.
        """
        iee_config: List[Dict[str, Any]] = config.get("key_blobs", [config.get("key_blob")])
        family = config["family"]
        master = config.get("master")
        if master:
            master = MasterId[master]
        ibkek1 = get_key(
            config.get(
                "ibkek1",
                "0x000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F",
            ),
            32,
        )
        ibkek2 = get_key(
            config.get(
                "ibkek2",
                "0x202122232425262728292A2B2C2D2E2F303132333435363738393A3B3C3D3E3F",
            ),
            32,
        )

        logger.debug(f"Loaded IBKEK1: {ibkek1.hex()}")
        logger.debug(f"Loaded IBKEK2: {ibkek2.hex()}")

        keyblob_address = value_to_int(config["keyblob_address"])
        start_address = min(
            [value_to_int(addr.get("start_address", 0xFFFFFFFF)) for addr in iee_config]
        )

        data_blobs: Optional[List[Dict]] = config.get("data_blobs")
        binaries = None
        if data_blobs:
            # start address to calculate offset from keyblob, min from keyblob or data blob address
            # pylint: disable-next=nested-min-max
            start_address = min(
                min([value_to_int(addr.get("address", 0xFFFFFFFF)) for addr in data_blobs]),
                start_address,
            )
            binaries = BinaryImage(
                filepath_from_config(
                    config, "encrypted_name", "encrypted_blobs", config_dir, config["output_folder"]
                ),
                offset=start_address - keyblob_address,
                alignment=IeeKeyBlob._ENCRYPTION_BLOCK_SIZE,
            )
            for data_blob in data_blobs:
                address = value_to_int(
                    data_blob.get("address", 0), keyblob_address + binaries.offset
                )

                binary = BinaryImage.load_binary_image(
                    path=data_blob["data"],
                    search_paths=search_paths,
                    offset=address - keyblob_address - binaries.offset,
                    alignment=IeeKeyBlob._ENCRYPTION_BLOCK_SIZE,
                    size=0,
                )

                binaries.add_image(binary)

        iee = IeeNxp(family, keyblob_address, ibkek1, ibkek2, binaries=binaries)

        for key_blob_cfg in iee_config:
            aes_mode = key_blob_cfg["aes_mode"]
            region_lock = "LOCK" if key_blob_cfg.get("region_lock") else "UNLOCK"
            key_size = key_blob_cfg["key_size"]

            attributes = IeeKeyBlobAttribute(
                IeeKeyBlobLockAttributes[region_lock],
                IeeKeyBlobKeyAttributes[key_size],
                IeeKeyBlobModeAttributes[aes_mode],
            )

            key1 = get_key(key_blob_cfg["key1"], attributes.key1_size)
            key2 = get_key(key_blob_cfg["key2"], attributes.key2_size)

            start_addr = value_to_int(key_blob_cfg.get("start_address", start_address))
            end_addr = value_to_int(key_blob_cfg.get("end_address", 0xFFFFFFFF))
            page_offset = value_to_int(key_blob_cfg.get("page_offset", 0))

            iee.add_key_blob(
                IeeKeyBlob(
                    attributes=attributes,
                    start_addr=start_addr,
                    end_addr=end_addr,
                    key1=key1,
                    key2=key2,
                    page_offset=page_offset,
                    master=master,
                )
            )

        return iee



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Image."""

from datetime import datetime, timezone
from io import SEEK_CUR, SEEK_END, BufferedReader, BytesIO
from struct import unpack_from
from typing import Any, List, Optional, Tuple, Union

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers.aead import AESCCM
from cryptography.hazmat.primitives.serialization import Encoding

from spsdk.exceptions import SPSDKError, SPSDKParsingError
from spsdk.utils.crypto import crypto_backend
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import DebugInfo, align, align_block, extend_block

from .commands import (
    CmdAuthData,
    CmdInstallKey,
    EnumAlgorithm,
    EnumAuthDat,
    EnumCertFormat,
    EnumEngine,
    EnumInsKey,
)
from .header import Header, Header2
from .misc import NotEnoughBytesException, read_raw_data, read_raw_segment
from .secret import MAC, CertificateImg, Signature, SrkTable
from .segments import (
    AbstractFCB,
    FlexSPIConfBlockFCB,
    PaddingFCB,
    SegAPP,
    SegBDS3a,
    SegBDS3b,
    SegBDT,
    SegBEE,
    SegBIC1,
    SegCSF,
    SegDCD,
    SegIVT2,
    SegIVT3a,
    SegIVT3b,
    SegTag,
    SegXMCD,
    XMCDHeader,
)

########################################################################################################################
# i.MX Boot Image Classes
########################################################################################################################


class EnumAppType(Enum):
    """Type of the application image."""

    SCFW = 1
    M4_0 = 2
    M4_1 = 3
    APP = 4  # actually this means APP or A35 or A53
    A72 = 5
    SCD = 6


class BootImgBase:
    """IMX Boot Image Base."""

    def __init__(self, address: int, offset: int) -> None:
        """Initialize boot image object.

        :param address: The start address of img in target memory
        :param offset: The IVT offset
        """
        self.address = address
        self.offset = offset
        self._dcd: Optional[SegDCD] = None

    @property
    def dcd(self) -> Optional[SegDCD]:
        """Device configuration data (DCD) segment; None if not assigned."""
        return self._dcd

    @dcd.setter
    def dcd(self, value: SegDCD) -> None:
        """Setter.

        :param value: new DCD segment
        """
        assert isinstance(value, SegDCD)
        self._dcd = value

    def info(self) -> str:
        """Text info about the instance.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def add_image(self, data: bytes, img_type: EnumAppType, address: int) -> None:
        """Add specific image into the main boot image.

        :param data: Raw binary data of the application image
        :param img_type: see EnumAppType
        :param address: TBD
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def export(self) -> bytes:
        """Binary representation of the instance (serialization).

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> "BootImgBase":
        """Parse of IMX Boot Image Base.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")


########################################################################################################################
# Boot Image V1 Segments (i.MX5)
########################################################################################################################

# Obsolete, will not be implemented

########################################################################################################################
# Boot Image V2 (i.MX-RT)
########################################################################################################################


# pylint: disable=too-many-public-methods
class BootImgRT(BootImgBase):
    """IMX Boot Image v2."""

    # offset of the BEE PRDB Header segment
    BEE_OFFSET = 0x400
    # IVT offset for NOR flash
    IVT_OFFSET_NOR_FLASH = 0x1000
    # IVT offset for other memories
    IVT_OFFSET_OTHER = 0x400
    # IVT offset for other memories
    IVT_OFFSET_OTHER2 = 0xC00
    # supported IVT offsets
    IVT_OFFSETS = (0, IVT_OFFSET_OTHER, IVT_OFFSET_OTHER2, IVT_OFFSET_NOR_FLASH)
    # possible FCB offsets
    FCB_OFFSETS = (0, 0x400)
    # XMCD offset relative to IVT
    XMCD_IVT_OFFSET = 0x40
    # list of supported versions
    VERSIONS = (0x40, 0x41, 0x42, 0x43)
    # The offset and align value of APP segment (for XIP and non-XIP image)
    XIP_APP_OFFSET = 0x2000
    NON_XIP_APP_OFFSET = 0x1000
    # The value of CSF segment size
    CSF_SIZE = 0x2000
    # The length of BDT segment
    BDT_SIZE = 0x20
    # The length of DEK key section; Note: Dek key is just 16 bytes
    DEK_SIZE = 0x200  # Is this sector size alignment???

    def __init__(
        self,
        address: int,
        offset: int = IVT_OFFSET_NOR_FLASH,
        version: int = 0x40,
        plugin: bool = False,
    ):
        """Initialize boot image object.

        :param address: The start address of img in target memory, where the image is executed
        :param offset: The IVT offset; use IVT_OFFSET_NOR_FLASH for NOR-FLASH or IVT_OFFSET_OTHER
        :param version: The version of boot img format; default value should be used
        :param plugin: Do not use; see `self.plugin` property
        :raises SPSDKError: If invalid IVT offset
        :raises SPSDKError: If invalid version
        :raises SPSDKError: If Plugin is not supported
        """
        if offset not in BootImgRT.IVT_OFFSETS:
            raise SPSDKError("Invalid IVT offset")
        if version not in self.VERSIONS:
            raise SPSDKError("Invalid version")
        if plugin is True:
            raise SPSDKError("Plugin is not supported")  # not supported yet
        super().__init__(address, offset)
        self._nonce: Optional[bytes] = None
        self._dek_key: Optional[bytes] = None
        self._mac: Optional[bytes] = None
        self._fcb: AbstractFCB = PaddingFCB(self.IVT_OFFSET_OTHER)
        self._bee: SegBEE = SegBEE([])
        self._ivt: SegIVT2 = SegIVT2(version)
        self._bdt: SegBDT = SegBDT(plugin=int(plugin))
        self._app: SegAPP = SegAPP()
        self._dcd: Optional[SegDCD] = None
        self._csf: Optional[SegCSF] = None
        self._xmcd: Optional[SegXMCD] = None

    @property
    def version(self) -> int:
        """Version of the image format; must be from BootImgRT.VERSIONS."""
        return self._ivt.version

    @property
    def dek_key(self) -> Optional[bytes]:
        """DEK key for encrypted images; None for non-encrypted images."""
        return self._dek_key

    @dek_key.setter
    def dek_key(self, value: bytes) -> None:
        """Setter.

        :param value: DEK key for encrypted images
        :raises SPSDKError: If invalid length of DEK key
        """
        if len(value) != MAC.AES128_BLK_LEN:
            raise SPSDKError("Invalid length of DEK key")
        self._dek_key = value

    @property
    def plugin(self) -> bool:
        """Flag whether it is plugin image type; It is not fully supported by SPSDK yet.

        Plugin is designed to load a boot image from devices that are not natively supported by boot ROM.
        """
        return bool(self._bdt.plugin)

    @property
    def ivt(self) -> SegIVT2:
        """Image Vector Table (IVT) segment."""
        return self._ivt

    @ivt.setter
    def ivt(self, value: SegIVT2) -> None:
        """Setter.

        :param value: new value
        """
        assert isinstance(value, SegIVT2)
        self._ivt = value

    @property
    def ivt_offset(self) -> int:
        """Offset of the Image Vector Table (IVT) in the image."""
        return self.offset

    @ivt_offset.setter
    def ivt_offset(self, value: int) -> None:
        """Setter.

        :param value: new IVT offset
        :raises SPSDKError: If invalid IVT offset
        """
        if value not in self.IVT_OFFSETS:
            raise SPSDKError("Invalid IVT offset")
        self.offset = value

    @property
    def bdt(self) -> SegBDT:
        """Boot Data Table."""
        return self._bdt

    @bdt.setter
    def bdt(self, value: SegBDT) -> None:
        """Setter.

        :param value: new BDT value
        """
        assert isinstance(value, SegBDT)
        self._bdt = value

    @property
    def app(self) -> SegAPP:
        """Segment with application image."""
        return self._app

    @app.setter
    def app(self, value: SegAPP) -> None:
        """Setter.

        :param value: new application image
        """
        assert isinstance(value, SegAPP)
        self._app = value

    @property
    def csf(self) -> Optional[SegCSF]:
        """Command Sequence File (CSF), signature block for Secure Boot."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        """Setter.

        :param value: new CSF
        """
        assert isinstance(value, SegCSF)
        self._csf = value
        self._update()

    @property
    def enabled_csf(self) -> Optional[SegCSF]:
        """Enabled Command Sequence File (CSF) segment; None if CSF is not defined or it is not enabled."""
        return None if (self.csf is None) or not self.csf.enabled else self.csf

    @property
    def fcb(self) -> AbstractFCB:
        """Flash Configuration(Control) Block, binary data; content depends on FLASH type."""
        return self._fcb

    @fcb.setter
    def fcb(self, fcb: AbstractFCB) -> None:
        """Setter.

        :param fcb: FCB instance to be set
        """
        assert isinstance(fcb, AbstractFCB)
        self._fcb = fcb

    def set_flexspi_fcb(self, data: Union[bytes, FlexSPIConfBlockFCB]) -> None:
        """Set FlexSPI external FLASH configuration.

        :param data: FlexSPIConfBlockFCB or binary data representing
        """
        self.fcb = (
            data if isinstance(data, FlexSPIConfBlockFCB) else FlexSPIConfBlockFCB.parse(data)
        )

    @property
    def xmcd(self) -> Optional[SegXMCD]:
        """Return the XMCD block."""
        return self._xmcd

    @xmcd.setter
    def xmcd(self, xmcd: SegXMCD) -> None:
        """Sets the XMCD block."""
        assert isinstance(xmcd, SegXMCD)
        self._xmcd = xmcd

    def set_xmcd(self, data: Union[bytes, SegXMCD]) -> None:
        """Sets the XMCD block."""
        self.xmcd = data if isinstance(data, SegXMCD) else SegXMCD.parse(data)

    @property
    def bee(self) -> SegBEE:
        """:return: BEE segment that contains configuration of encrypted XIP.

        By default, BEE segment is empty. PRDB regions may be specified only for XIP images.
        """
        return self._bee

    @bee.setter
    def bee(self, bee: SegBEE) -> None:
        """Setter.

        :param bee: BEE instance to be set
        """
        assert isinstance(bee, SegBEE)
        self._bee = bee

    @property
    def app_offset(self) -> int:
        """:return: offset in the binary image, where the application starts.

        Please mind: the offset include FCB block (even the FCB block is not exported)
        The offset is 0x2000 for XIP images and 0x1000 for non-XIP images
        """
        return self.get_app_offset(self.ivt_offset)

    @staticmethod
    def get_app_offset(ivt_offset: int) -> int:
        """:return: offset in the binary image, where the application starts.

        :param ivt_offset: Offset of IVT segment

        Please mind: the offset include FCB block (even the FCB block is not exported)
        The offset is 0x2000 for XIP images and 0x1000 for non-XIP images
        """
        return (
            BootImgRT.XIP_APP_OFFSET
            if (ivt_offset == BootImgRT.IVT_OFFSET_NOR_FLASH)
            else BootImgRT.NON_XIP_APP_OFFSET
        )

    @property
    def size(self) -> int:
        """Size of the exported binary data.

        Please mind, FCB is exported optionally, but it is always included in the size
        """
        if self.fcb.enabled:
            result = self.app_offset + self.app.space
        else:
            result = self.app_offset + self.app.space - self.ivt_offset
        if (self.csf is not None) and self.csf.enabled:
            result += self.csf.space
        return result

    def _update(self) -> None:
        """Update Image Object."""
        # fcb
        self.fcb.padding_len = self.BEE_OFFSET - self.fcb.size if self.fcb.enabled else 0
        # bee
        if (self.ivt_offset == self.IVT_OFFSET_NOR_FLASH) and self.fcb.enabled:
            self.bee.padding_len = self.ivt_offset - self.BEE_OFFSET - self.bee.size
        else:
            self.bee.padding_len = 0
        self.bee.update()
        # padding for APP sections
        self.app.padding_len = 0
        # Set IVT section
        self.ivt.padding_len = 0
        self.ivt.ivt_address = self.address + self.ivt_offset
        self.ivt.bdt_address = self.ivt.ivt_address + self.ivt.space
        self.ivt.dcd_address = 0
        self.ivt.csf_address = 0
        # Set BDT section
        self.bdt.app_start = self.address
        self.bdt.app_length = self.app_offset + self.app.size
        self.bdt.plugin = 1 if self.plugin else 0
        self.bdt.padding_len = self.BDT_SIZE - self.bdt.size
        if self.dcd is not None:
            self.ivt.dcd_address = self.ivt.bdt_address + self.bdt.space
            self.dcd.padding_len = 0
        csf = self.enabled_csf
        if csf:
            self._update_csf(csf)

    @property
    def dek_ram_address(self) -> int:
        """Address of the DEK key in the RAM memory retrieved from the corresponding command.

        -1 if the image does not contain command for DEK key installation
        """
        csf = self.enabled_csf
        if csf:
            for cmd in csf.commands:
                if isinstance(cmd, CmdInstallKey) and (
                    cmd.certificate_format == EnumCertFormat.BLOB
                ):
                    return cmd.cmd_data_location
        return -1

    @property
    def dek_img_offset(self) -> int:
        """Offset of the DEK key in the image; -1 if DEK key address is available (see `dek_ram_address`)."""
        result = self.dek_ram_address
        return result if result < 0 else result - self.address

    def _update_csf(self, csf: SegCSF) -> None:
        """Update CSF segment.

        :param csf: CSF segment tu be updated
        :raises SPSDKError: If nonce not present
        :raises SPSDKError: If mac not present
        """
        self.app.padding_len = align(self.app.size, 0x1000) - self.app.size
        csf.update(True)
        self.ivt.csf_address = self.address + self.app_offset + self.app.space
        csf.padding_len = self.CSF_SIZE - csf.size
        self.bdt.app_length = self.app_offset + self.app.space + csf.space
        if self.hab_encrypted:
            # calculate address of a DEK key
            for cmd in csf.commands:
                if isinstance(cmd, CmdInstallKey) and (
                    cmd.certificate_format == EnumCertFormat.BLOB
                ):
                    cmd.cmd_data_location = self.address + self.bdt.app_length
            #
            self.bdt.app_length += self.DEK_SIZE  # to include DEK
            # update encryption signature
            if not self._nonce:
                raise SPSDKError("Nonce not present")
            if not self._mac:
                raise SPSDKError("Mac not present")
            for mac in csf.macs:
                mac.update_aead_encryption_params(self._nonce, self._mac)

    def info(self) -> str:
        """Text info about the instance."""
        self._update()
        # Print FCB
        msg = "#" * 60 + "\n"
        msg += "# FCB (Flash Configuration Block)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.fcb.info()
        # Print BEE
        if self.bee_encrypted:
            msg += "#" * 60 + "\n"
            msg += "# BEE (Encrypted XIP configuration)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.bee.info()
        # Print IVT
        msg += "#" * 60 + "\n"
        msg += "# IVT (Image Vector Table)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.ivt.info()
        # Print BDI
        msg += "#" * 60 + "\n"
        msg += "# BDI (Boot Data Info)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.bdt.info()
        # Print DCD
        if (self.dcd is not None) and self.dcd.enabled:
            msg += "#" * 60 + "\n"
            msg += "# DCD (Device Config Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.dcd.info()
        # Print XMCD
        if self.xmcd:
            msg += "#" * 60 + "\n"
            msg += "# XMCD (External Memory Configuration Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.xmcd.info()
        # Print CSF
        csf = self.enabled_csf
        if csf:
            msg += "#" * 60 + "\n"
            msg += "# CSF (Code Signing Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += csf.info()
        return msg

    def add_image(
        self,
        data: bytes,
        img_type: EnumAppType = EnumAppType.APP,
        address: int = -1,
        dek_key: Optional[bytes] = None,
        nonce: Optional[bytes] = None,
    ) -> None:
        """Add specific image into the main boot image.

        :param data: Raw data of img
        :param img_type: value must be EnumAppType.APP, no other options supported in this class
        :param address: start address of the application (entry point); Use -1 to detect the address from the image
        :param dek_key: key for AES128 image HAB encryption [16 bytes],
                    - use None for non-encrypted images;
                    - use empty bytes to create random key (recommended)
                    - use fixed key for testing to produce stable output
        :param nonce: initial vector for AEAD HAB encryption, if not specified random value is used;
                        For non-encrypted image use `None`
                        The parameter should be used only for testing to produce stable output
        :raises ValueError: if any parameter is not valid
        :raises SPSDKError: If invalid image type
        :raises SPSDKError: If image was already added
        :raises SPSDKError: If entry_addr not detected from image, must be specified explicitly
        :raises SPSDKError: If hab is not encrypted
        :raises SPSDKError: If nonce is not empty
        """
        if img_type != EnumAppType.APP:
            raise SPSDKError("Invalid image type")
        if self.app.data:
            raise SPSDKError("Image was already added")
        entry_addr = unpack_from("<I", data, 4)[0]
        if entry_addr == 0:  # there can be padding for images located in RAM, see flashloader
            entry_addr = address
            if entry_addr <= 0:
                raise SPSDKError("entry_addr not detected from image, must be specified explicitly")
        elif address >= 0 and address != entry_addr:
            raise SPSDKError("entry_address does not match with the image")
        self._ivt.app_address = entry_addr
        self.app.data = data
        if dek_key is not None:  # encrypted?
            # initialize DEK key
            self._dek_key = bytes([0]) * MAC.AES128_BLK_LEN if len(dek_key) == 0 else dek_key
            if len(self._dek_key) != MAC.AES128_BLK_LEN:
                raise SPSDKError(f"Invalid dek_key length, expected {MAC.AES128_BLK_LEN} bytes")
            # initialize NONCE
            if nonce:
                self._nonce = nonce
            nonce_len = BootImgRT.aead_nonce_len(self.app.size)
            if self._nonce is None:
                self._nonce = crypto_backend().random_bytes(nonce_len)
            elif len(self._nonce) != nonce_len:
                raise SPSDKError(f"Invalid nonce length, expected {nonce_len} bytes")
            # encrypt APP
            if not self.hab_encrypted:
                raise SPSDKError("Hab is not encrypted")
            self.app.data = self._hab_encrypt_app_data(align_block(data, MAC.AES128_BLK_LEN))
        else:
            if nonce is not None:
                raise SPSDKError("Nonce is not empty")

    def add_dcd_bin(self, data: bytes) -> None:
        """Add DCD binary data.

        :param data: DCD binary data to be added
        :raises SPSDKError: If DCD is already present
        :raises SPSDKError: If DCD is not enabled
        """
        if self.dcd is not None:
            raise SPSDKError("DCD is already present")
        self.dcd = SegDCD.parse(data)
        if not self.dcd:
            raise SPSDKError("DCD must be enabled to include DCD into export")

    def add_csf_standard_auth(
        self,
        version: int,
        srk_table: SrkTable,
        src_key_index: int,
        csf_cert: bytes,
        csf_priv_key: bytes,
        img_cert: bytes,
        img_priv_key: bytes,
    ) -> None:
        """Add CSF with standard authentication.

        Before calling, application image and address must be assigned

        :param version: CSF segment version
        :param srk_table: SRK table of root certificates; must contain min 1, max 4 certificates
        :param src_key_index: index of selected SRK key used for authentication
        :param csf_cert: CSF certificate
        :param csf_priv_key: CSF private key; decrypted binary data in PEM format
        :param img_cert: IMG certificate
        :param img_priv_key: IMG private key; decrypted binary data in PEM format
        :raises SPSDKError: If invalid length of srk table
        :raises SPSDKError: If invalid index of selected SRK key
        :raises SPSDKError: If application data not present
        """
        if not 1 <= len(srk_table) <= 4:
            raise SPSDKError("Invalid length of srk table")
        if not 0 <= src_key_index < len(srk_table):
            raise SPSDKError("Invalid index of selected SRK key")
        csf = SegCSF(version=version, enabled=True)
        # install SRK
        cmd_ins = CmdInstallKey(
            EnumInsKey.CLR, EnumCertFormat.SRK, EnumAlgorithm.SHA256, src_key_index, 0
        )
        cmd_ins.cmd_data_reference = srk_table
        csf.append_command(cmd_ins)
        # install CSF certificate
        cmd_ins = CmdInstallKey(EnumInsKey.CSF, EnumCertFormat.X509, EnumAlgorithm.ANY, 0, 1)
        cert = x509.load_pem_x509_certificate(csf_cert, default_backend())
        cmd_ins.cmd_data_reference = CertificateImg(
            version=version, data=cert.public_bytes(Encoding.DER)
        )
        csf.append_command(cmd_ins)
        # authenticate content of the CSF segment
        cmd_auth = CmdAuthData(
            EnumAuthDat.CLR,
            1,
            EnumCertFormat.CMS,
            EnumEngine.DCP,
            certificate=cert,
            private_key_pem_data=csf_priv_key,
        )
        cmd_auth.cmd_data_reference = Signature(version=version)
        csf.append_command(cmd_auth)
        # install image certificate
        cmd_ins = CmdInstallKey(EnumInsKey.CLR, EnumCertFormat.X509, EnumAlgorithm.ANY, 0, 2)
        cert = x509.load_pem_x509_certificate(img_cert, default_backend())
        cmd_ins.cmd_data_reference = CertificateImg(
            version=version, data=cert.public_bytes(Encoding.DER)
        )
        csf.append_command(cmd_ins)
        # authenticate image data
        cmd_auth = CmdAuthData(
            EnumAuthDat.CLR,
            2,
            EnumCertFormat.CMS,
            EnumEngine.DCP,
            certificate=cert,
            private_key_pem_data=img_priv_key,
        )
        cmd_auth.append(self.address + self.ivt_offset, SegIVT2.SIZE + BootImgRT.BDT_SIZE)
        if self.dcd:
            cmd_auth.append(
                self.address + self.ivt_offset + SegIVT2.SIZE + BootImgRT.BDT_SIZE,
                self.dcd.size,
            )
        app_data = self.app.data
        if app_data is None:
            raise SPSDKError("Application data not present")
        cmd_auth.append(self.address + self.app_offset, align(len(app_data), 16))
        cmd_auth.cmd_data_reference = Signature(version=version)
        csf.append_command(cmd_auth)
        self.csf = csf

    @property
    def bee_encrypted(self) -> bool:
        """True if BEE encrypted XIP image (with SW keys); False otherwise; see also `hab_encrypted`."""
        return self.bee.size > 0

    @property
    def hab_encrypted(self) -> bool:
        """True if HAB encrypted; False otherwise; see also `bee_encrypted`."""
        return self._dek_key is not None

    @staticmethod
    def aead_nonce_len(app_data_len: int) -> int:
        """Nonce len for AEAD encryption.

        Note: The code was taken from CST tool
        """
        if app_data_len < 0x10000:
            len_bytes = 2
        elif app_data_len < 0x1000000:
            len_bytes = 3
        else:
            len_bytes = 4
        return 16 - 1 - len_bytes  # AES_BLOCK_BYTES - FLAG_BYTES - len_bytes

    def _hab_encrypt_app_data(self, app_data: bytes) -> bytes:
        """HAB Encrypt application data.

        :param app_data: application data to be encrypted
        :return: encrypted application data (using HAB encryption)
        :raises SPSDKError: If nonce is not present
        :raises SPSDKError: If invalid length of application data
        :raises SPSDKError: If DEK key is not present
        :raises SPSDKError: If invalid length of encrypted data
        """
        if self._nonce is None:
            raise SPSDKError("Nonce is not present")
        if not len(app_data) & (MAC.AES128_BLK_LEN - 1) == 0:
            raise SPSDKError("Invalid length of application data")
        dek = self.dek_key
        if dek is None:
            raise SPSDKError("DEK key is not present")
        aesccm = AESCCM(dek, tag_length=MAC.AES128_BLK_LEN)
        encr = aesccm.encrypt(self._nonce, app_data, b"")
        if len(encr) != len(app_data) + 16:
            raise SPSDKError("Invalid length of encrypted data")
        self._mac = encr[-16:]
        return encr[:-16]

    @property
    def decrypted_app_data(self) -> bytes:
        """Return decrypted binary application data.

        Note: dek key, mac and nonce must be assigned for decryption
        :raises SPSDKError: If application not present
        :raises SPSDKError: If invalid length of application data
        :raises SPSDKError: If Mac or nonce or dek not present
        """
        app_data = self.app.data
        if not app_data:
            raise SPSDKError("Application not present")
        if not self.hab_encrypted:
            return app_data

        if not len(app_data) & (MAC.AES128_BLK_LEN - 1) == 0:
            raise SPSDKError("Invalid length of application data")
        mac = self._mac
        dek = self.dek_key
        if not (mac and self._nonce and dek):
            raise SPSDKError("Mac or nonce or dek not present")
        aesccm = AESCCM(dek, tag_length=MAC.AES128_BLK_LEN)
        res = aesccm.decrypt(self._nonce, app_data + mac, b"")
        return res

    def add_csf_encrypted(
        self,
        version: int,
        srk_table: SrkTable,
        src_key_index: int,
        csf_cert: bytes,
        csf_priv_key: bytes,
        img_cert: bytes,
        img_priv_key: bytes,
    ) -> None:
        """Add CSF with image encryption.

        Before calling, application image and address must be assigned

        :param version: CSF segment version
        :param srk_table: SRK table of root certificates; must contain min 1, max 4 certificates
        :param src_key_index: index of selected SRK key used for authentication, 0..srk_table.len - 1
        :param csf_cert: CSF certificate
        :param csf_priv_key: CSF private key; decrypted binary data in PEM format
        :param img_cert: IMG certificate
        :param img_priv_key: IMG private key; decrypted binary data in PEM format
        :raises SPSDKError: If invalid length of srk table
        :raises SPSDKError: If invalid index of srk table
        :raises SPSDKError: If application data is not present
        """
        if not 1 <= len(srk_table) <= 4:
            raise SPSDKError("Invalid length of srk table")
        if not 0 <= src_key_index < len(srk_table):
            raise SPSDKError("Invalid index of srk table")
        csf = SegCSF(version=version, enabled=True)
        # install SRK
        cmd_ins = CmdInstallKey(
            EnumInsKey.CLR, EnumCertFormat.SRK, EnumAlgorithm.SHA256, src_key_index, 0
        )
        cmd_ins.cmd_data_reference = srk_table
        csf.append_command(cmd_ins)
        # install CSF certificate
        cmd_ins = CmdInstallKey(EnumInsKey.CSF, EnumCertFormat.X509, EnumAlgorithm.ANY, 0, 1)
        cert = x509.load_pem_x509_certificate(csf_cert, default_backend())
        cmd_ins.cmd_data_reference = CertificateImg(
            version=version, data=cert.public_bytes(Encoding.DER)
        )
        csf.append_command(cmd_ins)
        # authenticate content of the CSF segment
        cmd_auth = CmdAuthData(
            EnumAuthDat.CLR,
            1,
            EnumCertFormat.CMS,
            EnumEngine.DCP,
            certificate=cert,
            private_key_pem_data=csf_priv_key,
        )
        cmd_auth.cmd_data_reference = Signature(version=version)
        csf.append_command(cmd_auth)
        # install image certificate
        cmd_ins = CmdInstallKey(EnumInsKey.CLR, EnumCertFormat.X509, EnumAlgorithm.ANY, 0, 2)
        cert = x509.load_pem_x509_certificate(img_cert, default_backend())
        cmd_ins.cmd_data_reference = CertificateImg(
            version=version, data=cert.public_bytes(Encoding.DER)
        )
        csf.append_command(cmd_ins)
        # authenticate image data
        cmd_auth = CmdAuthData(
            EnumAuthDat.CLR,
            2,
            EnumCertFormat.CMS,
            EnumEngine.DCP,
            certificate=cert,
            private_key_pem_data=img_priv_key,
        )
        cmd_auth.append(self.address + self.ivt_offset, SegIVT2.SIZE + BootImgRT.BDT_SIZE)
        app_data = self.app.data
        if app_data is None:
            raise SPSDKError("Application data is not present")
        cmd_auth.cmd_data_reference = Signature(version=version)
        csf.append_command(cmd_auth)
        # install DEK key
        cmd_ins = CmdInstallKey(EnumInsKey.ABS, EnumCertFormat.BLOB, EnumAlgorithm.ANY, 0, 0)
        csf.append_command(cmd_ins)
        # check encrypted data
        cmd_auth = CmdAuthData(
            EnumAuthDat.CLR,
            0,
            EnumCertFormat.AEAD,
            EnumEngine.DCP,
            certificate=cert,
            private_key_pem_data=img_priv_key,
        )
        if app_data is None:
            raise SPSDKError("Application data is not present")
        cmd_auth.append(self.address + self.app_offset, align(len(app_data), 16))
        cmd_auth.cmd_data_reference = MAC(version=version, nonce_len=0xD, mac_len=16)
        csf.append_command(cmd_auth)
        #
        self.csf = csf

    def export_fcb(self, dbg_info: DebugInfo) -> bytes:
        """Export FCB segment.

        :param dbg_info: optional instance to provide info about exported data
        :return: binary FCB segment
        :raises SPSDKError: If invalid length of data
        """
        if not self.fcb.enabled:
            return b""
        data = self.fcb.export(dbg_info=dbg_info)
        if len(data) != self.fcb.space:
            raise SPSDKError("Invalid length of data")
        return data

    def export_bee(self, dbg_info: DebugInfo) -> bytes:
        """Export BEE segment.

        :param dbg_info: optional instance to provide info about exported data
        :return: binary BEE segment
        :raises SPSDKError: if any BEE region is configured for images not located in the FLASH
        """
        data = b""
        if self.ivt_offset == self.IVT_OFFSET_NOR_FLASH:
            data = self.bee.export(dbg_info=dbg_info)
        elif self.bee.space > 0:
            raise SPSDKError("BEE can be configured only for XIP images located in FLASH")
        return data

    def export_dcd(self, dbg_info: DebugInfo) -> bytes:
        """Export DCD segment.

        :param dbg_info: optional instance to provide info about exported data
        :return: binary DCD segment
        :raises SPSDKError: If DCD padding is not set
        """
        dcd_data = b""
        if (self.dcd is not None) and self.dcd.enabled:
            if self.dcd.padding_len != 0:
                raise SPSDKError("Padding can not be present")
            dcd_data = self.dcd.export()
            dbg_info.append_binary_section("DCD", dcd_data)
        return dcd_data

    def export_csf(
        self, dbg_info: DebugInfo, data: bytes, zulu: datetime = datetime.now(timezone.utc)
    ) -> bytes:
        """Export CSF segment.

        :param dbg_info: optional instance to provide info about exported data
        :param data: generated binary data used for creating of signature
        :param zulu: current UTC datetime
        :return: binary CFD segment
        """
        csf_data = b""
        if self.enabled_csf:
            dbg_info.append_section("CSF")
            base_data_addr = self.address if self.fcb.enabled else self.address + self.ivt_offset
            self.enabled_csf.update_signatures(zulu, data, base_data_addr)
            csf_data = self.enabled_csf.export(dbg_info=dbg_info)
        return csf_data

    def _bee_encrypt_img_data(self, data: bytes) -> bytes:
        """Encrypt data located in BEE regions.

        :param data: image data (including IVT offset) to be encrypted
        :return: the image with encrypted regions
        :raises SPSDKError: If image configuration is invalid and BEE encryption cannot be applied
        """
        if not self.bee_encrypted:
            return data

        if self.ivt_offset != self.IVT_OFFSET_NOR_FLASH:
            raise SPSDKError("BEE encryption is supported only for NOR FLASH")
        if self.hab_encrypted:
            raise SPSDKError("BEE encryption cannot be used for HAB encrypted images")

        # encrypt
        return data[: self.ivt_offset] + self.bee.encrypt_data(
            self.address + self.ivt_offset, data[self.ivt_offset :]
        )

    def export(
        self,
        zulu: datetime = datetime.now(timezone.utc),
        dbg_info: DebugInfo = DebugInfo.disabled(),
    ) -> bytes:
        """Export image as bytes array.

        :param zulu: optional UTC datetime; should be used only if you need fixed datetime for the test
                Note: the parameter is applied to CSF only, so it is not used for unsigned images
        :param dbg_info: optional instance to provide info about exported data
        :raises SPSDKError: If the image is not encrypted
        :raises SPSDKError: If padding is present
        :raises SPSDKError: If invalid alignment of application
        :return: bytes
        """
        csf = self.enabled_csf
        if csf:
            csf.update_signatures(zulu, b"", 0)  # dummy call to provide size of the CSF section
        elif self.dek_key is not None:
            raise SPSDKError("CSF must be assigned for encrypted images")

        self._update()
        dbg_info.append_section("RT10xxBootableImage")
        # FCB
        data = self.export_fcb(dbg_info)
        # BEE
        bee_data = self.export_bee(dbg_info)
        data += bee_data
        # IVT
        ivt_data = self.ivt.export()
        data += ivt_data
        dbg_info.append_binary_section("IVT", ivt_data)
        # BDT
        bdt_data = self.bdt.export()
        data += bdt_data
        dbg_info.append_binary_section("BDT", bdt_data)
        # DCD
        dcd_data = self.export_dcd(dbg_info)
        data += dcd_data
        # padding before APP
        app_alignment = self.app_offset if self.fcb.enabled else self.app_offset - self.ivt_offset
        if not app_alignment >= len(data):
            raise SPSDKError("Invalid alignment of application")
        data = extend_block(data, app_alignment)
        # APP
        app_data = self.app.export()
        data += app_data
        dbg_info.append_binary_section("APP", app_data)
        # CSF
        csf_data = self.export_csf(dbg_info, data=data, zulu=zulu)
        data += csf_data
        return self._bee_encrypt_img_data(data)

    @classmethod
    def _find_ivt_pos(
        cls, strm: Union[BufferedReader, BytesIO], size: Optional[int] = None
    ) -> Tuple[Header, int, int]:
        """Search IVT start position in the image; used by parser.

        :param strm: of image data; start seeking from current position
        :param size: maximum length
        :raises SPSDKError: Raised when IVT is not found
        :return: tuple with: Header, start position, end position
        """
        start_pos = strm.tell()
        end_pos = strm.seek(0, SEEK_END)

        if size:
            end_pos = min(start_pos + size, end_pos)

        for ivt_ofs in cls.IVT_OFFSETS:
            if start_pos + ivt_ofs > end_pos:
                break
            strm.seek(start_pos + ivt_ofs)
            header_data = read_raw_data(strm, Header.SIZE, no_seek=True)
            try:
                header = Header.parse(header_data, required_tag=SegTag.IVT2)
                if (header.length == SegIVT2.SIZE) and (header.param in cls.VERSIONS):
                    return header, start_pos + ivt_ofs, end_pos
            except SPSDKParsingError:  # ignore different header tags
                pass

        raise SPSDKError("IVT not found")

    @classmethod
    def _find_fcb_pos(
        cls, stream: Union[BufferedReader, BytesIO], size: Optional[int] = None
    ) -> Optional[int]:
        """Search for FCB start position.

        :param stream: data to search through
        :param size: maximal size to search through; default: whole stream
        :return: Starting location of FCB, None if FCB is not found
        """
        start_pos = stream.seek(0)
        end_pos = stream.seek(0, SEEK_END)

        if size:
            end_pos = min(start_pos + size, end_pos)

        for possible_offset in cls.FCB_OFFSETS:
            current_pos = start_pos + possible_offset
            if current_pos > end_pos:
                break
            stream.seek(current_pos)
            data = read_raw_data(stream, len(FlexSPIConfBlockFCB.TAG))
            if data == FlexSPIConfBlockFCB.TAG:
                return current_pos

        return None

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0,
        size: Optional[int] = None,
    ) -> "BootImgRT":
        """Parse bootable RT image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step (this parameter is not used for RT)
        :param size: parsing size; None to parse till the end of the stream
        :raises SPSDKError: Raised when the value type is incorrect
        :return: BootImgRT object
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f' Not correct value type: "{type(stream)}" !')

        header, start_pos, end_pos = cls._find_ivt_pos(stream, size)

        obj = BootImgRT(0, version=header.param)
        img_size = end_pos - start_pos
        if start_pos > 0:
            obj.ivt_offset = start_pos

        # Parse IVT
        obj.ivt = SegIVT2.parse(read_raw_segment(stream, SegTag.IVT2))

        # Try to find XMCD segment
        stream.seek(start_pos + cls.XMCD_IVT_OFFSET)
        try:
            xmcd_header = XMCDHeader.parse(read_raw_data(stream, XMCDHeader.SIZE))
            xmcd_data = read_raw_data(stream, xmcd_header.config_data_size)
            obj.xmcd = SegXMCD(header=xmcd_header, config_data=xmcd_data)
        except SPSDKParsingError:
            # No XMCD found
            pass
        # Parse BDT
        stream.seek(start_pos + obj.ivt.bdt_address - obj.ivt.ivt_address)
        obj.bdt = SegBDT.parse(read_raw_data(stream, SegBDT.SIZE))
        obj.ivt_offset = obj.ivt.ivt_address - obj.bdt.app_start
        obj.address = obj.bdt.app_start
        # Parse DCD
        if obj.ivt.dcd_address:
            stream.seek(start_pos + obj.ivt.dcd_address - obj.ivt.ivt_address)
            obj.dcd = SegDCD.parse(read_raw_segment(stream, SegTag.DCD))
        # Parse APP
        if obj.ivt.csf_address > 0:
            app_size = obj.ivt.csf_address - obj.ivt.ivt_address - (obj.app_offset - obj.ivt_offset)
        else:
            app_size = img_size - (obj.app_offset - obj.ivt_offset)
        obj.app.data = read_raw_data(stream, app_size, obj.app_offset - obj.ivt_offset + start_pos)
        obj.app.padding = 0
        # Parse CSF
        if obj.ivt.csf_address:
            csf_start = start_pos + (obj.ivt.csf_address - obj.ivt.ivt_address)
            obj.csf = SegCSF.parse(read_raw_data(stream, cls.CSF_SIZE, csf_start))
            # detect encrypted image using MAC section
            mac = next(obj.csf.macs, None)
            if mac:
                obj._nonce = mac.nonce
                obj._mac = mac.mac
                obj._dek_key = bytes([0]) * MAC.AES128_BLK_LEN  # dek key is not known

        # Parse FCB
        fcb_size = FlexSPIConfBlockFCB().size
        fcb_position = cls._find_fcb_pos(stream)
        if fcb_position is not None:
            fcb_data = read_raw_data(stream, fcb_size, fcb_position)
            obj.set_flexspi_fcb(fcb_data)
        else:
            obj.fcb = PaddingFCB(fcb_size, enabled=True)

        return obj


########################################################################################################################
# Boot Image V2 (i.MX6, i.MX7)
########################################################################################################################


class BootImg2(BootImgBase):
    """IMX Boot Image v2."""

    # The value of CSF segment size
    CSF_SIZE = 0x2000
    # The align value of APP segment
    APP_ALIGN = 0x1000
    # The value of img head size
    #           offset | size
    HEAD_SIZE = {0x400: 0xC00, 0x100: 0x300}

    @property
    def version(self) -> int:
        """Version of IMX Boot Image v2."""
        return self._ivt.version

    @version.setter
    def version(self, value: int) -> None:
        self._ivt.version = value

    @property
    def plugin(self) -> bool:
        """Plugin."""
        return self._plg

    @plugin.setter
    def plugin(self, value: bool) -> None:
        assert isinstance(value, bool)
        self._plg = value

    @property
    def ivt(self) -> SegIVT2:
        """IVT."""
        return self._ivt

    @ivt.setter
    def ivt(self, value: SegIVT2) -> None:
        assert isinstance(value, SegIVT2)
        self._ivt = value

    @property
    def bdt(self) -> SegBDT:
        """BDT."""
        return self._bdt

    @bdt.setter
    def bdt(self, value: SegBDT) -> None:
        assert isinstance(value, SegBDT)
        self._bdt = value

    @property
    def app(self) -> SegAPP:
        """APP."""
        return self._app

    @app.setter
    def app(self, value: SegAPP) -> None:
        assert isinstance(value, SegAPP)
        self._app = value

    @property
    def csf(self) -> SegCSF:
        """CSF."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        assert isinstance(value, SegCSF)
        self._csf = value

    @property
    def size(self) -> int:
        """Size of IMX Boot Image v2.."""
        result = self.ivt.space
        result += self.bdt.space
        if self.dcd:
            result += self.dcd.space
        result += self.app.space
        result += self.csf.space
        return result

    def __init__(
        self,
        address: int = 0,
        offset: int = 0x400,
        version: int = 0x41,
        plugin: bool = False,
    ) -> None:
        """Initialize boot image object.

        :param address: The start address of img in target memory
        :param offset: The IVT offset
        :param version: The version of boot img format
        :param plugin: if plugin
        """
        super().__init__(address, offset)
        self._ivt = SegIVT2(version)
        self._bdt = SegBDT()
        self._app = SegAPP()
        self._dcd = SegDCD()
        self._csf = SegCSF()
        self._plg = plugin

    def _update(self) -> None:
        """Update Image Object."""
        # Set zero padding for IVT and BDT sections
        self.ivt.padding = 0
        self.bdt.padding = 0
        # Calculate padding for DCD, APP and CSF sections
        tmp_val = self.ivt.space + self.bdt.space
        if self.dcd:
            tmp_val += self.dcd.size
        head_size = 0xC00 if self.offset not in self.HEAD_SIZE else self.HEAD_SIZE[self.offset]
        if self.dcd:
            self.dcd.padding = head_size - tmp_val
        tmp_val = self.app.size % self.APP_ALIGN
        self.app.padding = self.APP_ALIGN - tmp_val if tmp_val > 0 else 0
        # Set IVT section
        self.ivt.ivt_address = self.address + self.offset
        self.ivt.bdt_address = self.ivt.ivt_address + self.ivt.space
        if self.dcd:
            self.ivt.dcd_address = self.ivt.bdt_address + self.bdt.space
            self.ivt.app_address = self.ivt.dcd_address + self.dcd.space
        else:
            self.ivt.dcd_address = 0
            self.ivt.app_address = self.ivt.bdt_address + self.bdt.space
        if self.csf.enabled:
            self.ivt.csf_address = self.ivt.app_address + self.app.space
            self.csf.padding = self.CSF_SIZE - self.csf.size
        else:
            self.ivt.csf_address = 0
        # Set BDT section
        self.bdt.app_start = self.ivt.ivt_address - self.offset
        self.bdt.app_length = self.size + self.offset
        self.bdt.plugin = 1 if self.plugin else 0

    def info(self) -> str:
        """String representation of the IMX Boot Image v2."""
        self._update()
        # Print IVT
        msg = "#" * 60 + "\n"
        msg += "# IVT (Image Vector Table)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.ivt.info()
        # Print DBI
        msg += "#" * 60 + "\n"
        msg += "# BDI (Boot Data Info)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.bdt.info()
        # Print DCD
        if self.dcd:
            msg += "#" * 60 + "\n"
            msg += "# DCD (Device Config Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.dcd.info()
        # Print CSF
        if self.csf.enabled:
            msg += "#" * 60 + "\n"
            msg += "# CSF (Code Signing Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.csf.info()
        return msg

    def add_image(
        self, data: bytes, img_type: EnumAppType = EnumAppType.APP, address: int = 0
    ) -> None:
        """Add specific image into the main boot image.

        :param data: Raw data of img
        :param img_type: Type of img
        :param address: address in RAM
        :raises Exception: Raised when the data type is unknown
        """
        if img_type == EnumAppType.APP:
            self.app.data = data
            if address != 0:
                self.address = address
        else:
            raise SPSDKError("Unknown data type !")

    def export(self) -> bytes:
        """Export image as bytes array.

        :return: bytes
        """
        self._update()
        data = self.ivt.export()
        data += self.bdt.export()
        if self.dcd:
            data += self.dcd.export()
        data += self.app.export()
        data += self.csf.export()
        return data

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> "BootImg2":
        """Parse image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step
        :param size: parsing size
        :raises SPSDKError: Raised when value type is incorrect
        :raises SPSDKError: Raised when there is not an i.MX Boot Image
        :return: BootImg2 object
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f"Not correct value type: {type(stream)} !")

        header = Header()
        start_index = stream.tell()
        last_index = stream.seek(0, SEEK_END)
        stream.seek(start_index)

        if size:
            last_index = min(start_index + size, last_index)

        imx_image = False
        while start_index < (last_index - Header.SIZE):
            header = Header.parse(read_raw_data(stream, Header.SIZE, no_seek=True))
            if (
                header.tag == SegTag.IVT2
                or header.length == SegIVT2.SIZE
                or header.param in (0x40, 0x41, 0x42, 0x43)
            ):
                imx_image = True
                break

            start_index = stream.seek(step, SEEK_CUR)

        if not imx_image:
            raise SPSDKError("Not an i.MX Boot Image!")

        obj = BootImg2()
        if header.param:
            obj.version = header.param

        img_size = last_index - start_index
        if start_index > 0:
            obj.offset = start_index

        # Parse IVT
        obj.ivt = SegIVT2.parse(read_raw_segment(stream, SegTag.IVT2))
        # Parse BDT
        obj.bdt = SegBDT.parse(read_raw_data(stream, SegBDT.SIZE))
        obj.offset = obj.ivt.ivt_address - obj.bdt.app_start
        obj.address = obj.bdt.app_start
        obj.plugin = bool(obj.bdt.plugin)
        # Parse DCD
        if obj.ivt.dcd_address:
            obj.dcd = SegDCD.parse(read_raw_segment(stream, SegTag.DCD))
            obj.dcd.padding = (obj.ivt.app_address - obj.ivt.dcd_address) - obj.dcd.size
        # Parse APP
        app_start = start_index + (obj.ivt.app_address - obj.ivt.ivt_address)
        app_size = (
            obj.ivt.csf_address - obj.ivt.app_address
            if obj.ivt.csf_address
            else obj.bdt.app_length - (obj.bdt.app_start - obj.ivt.app_address)
        )
        app_size = img_size - app_start if app_size > (img_size - app_start) else app_size
        obj.app.data = read_raw_data(stream, app_size, app_start)
        obj.app.padding = 0
        # Parse CSF
        if obj.ivt.csf_address:
            csf_start = start_index + (obj.ivt.csf_address - obj.ivt.ivt_address)
            try:
                obj.csf = SegCSF.parse(read_raw_data(stream, cls.CSF_SIZE, csf_start))
                obj.csf.padding = cls.CSF_SIZE - obj.csf.size
            except NotEnoughBytesException:
                pass

        return obj


########################################################################################################################
# Boot Image V2b (i.MX8M)
########################################################################################################################


class BootImg8m(BootImgBase):
    """IMX Boot Image."""

    # The value of CSF segment size
    CSF_SIZE = 0x2000
    # The align value of APP segment
    APP_ALIGN = 0x1000
    # The value of img head size
    #           offset | size
    HEAD_SIZE = {0x400: 0xC00, 0x100: 0x300}

    @property
    def version(self) -> int:
        """Version of IMX Boot Image."""
        return self._ivt.version

    @version.setter
    def version(self, value: int) -> None:
        self._ivt.version = value

    @property
    def plugin(self) -> bool:
        """Plugin."""
        return self._plg

    @plugin.setter
    def plugin(self, value: bool) -> None:
        assert isinstance(value, bool)
        self._plg = value

    @property
    def ivt(self) -> SegIVT2:
        """IVT."""
        return self._ivt

    @ivt.setter
    def ivt(self, value: SegIVT2) -> None:
        assert isinstance(value, SegIVT2)
        self._ivt = value

    @property
    def bdt(self) -> SegBDT:
        """BDT."""
        return self._bdt

    @bdt.setter
    def bdt(self, value: SegBDT) -> None:
        assert isinstance(value, SegBDT)
        self._bdt = value

    @property
    def app(self) -> SegAPP:
        """APP."""
        return self._app

    @app.setter
    def app(self, value: SegAPP) -> None:
        assert isinstance(value, SegAPP)
        self._app = value

    @property
    def csf(self) -> SegCSF:
        """CSF."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        assert isinstance(value, SegCSF)
        self._csf = value

    @property
    def size(self) -> int:
        """Size of IMX Boot Image."""
        result = self.ivt.space
        result += self.bdt.space
        if self.dcd:
            result += self.dcd.space
        result += self.app.space
        result += self.csf.space
        return result

    def __init__(
        self,
        address: int = 0,
        offset: int = 0x400,
        version: int = 0x41,
        plugin: bool = False,
    ) -> None:
        """Initialize boot image object.

        :param address: The start address of img in target memory
        :param offset: The IVT offset
        :param version: The version of boot img format
        :param plugin: if plugin
        """
        super().__init__(address, offset)
        self._ivt = SegIVT2(version)
        self._bdt = SegBDT()
        self._app = SegAPP()
        self._dcd = SegDCD()
        self._csf = SegCSF()
        self._plg = plugin

    def _update(self) -> None:
        # Set zero padding for IVT and BDT sections
        self.ivt.padding = 0
        self.bdt.padding = 0
        # Calculate padding for DCD, APP and CSF sections
        tmp_val = self.ivt.space + self.bdt.space
        if self.dcd:
            tmp_val += self.dcd.size
        head_size = 0xC00 if self.offset not in self.HEAD_SIZE else self.HEAD_SIZE[self.offset]
        if self.dcd:
            self.dcd.padding = head_size - tmp_val
        tmp_val = self.app.size % self.APP_ALIGN
        self.app.padding = self.APP_ALIGN - tmp_val if tmp_val > 0 else 0
        # Set IVT section
        self.ivt.ivt_address = self.address + self.offset
        self.ivt.bdt_address = self.ivt.ivt_address + self.ivt.space
        if self.dcd:
            if self.dcd.enabled:
                self.ivt.dcd_address = self.ivt.bdt_address + self.bdt.space
                self.ivt.app_address = self.ivt.dcd_address + self.dcd.space
            else:
                self.ivt.dcd_address = 0
                self.ivt.app_address = self.ivt.bdt_address + self.bdt.space
        if self.csf.enabled:
            self.ivt.csf_address = self.ivt.app_address + self.app.space
            self.csf.padding = self.CSF_SIZE - self.csf.size
        else:
            self.ivt.csf_address = 0
        # Set BDT section
        self.bdt.app_start = self.ivt.ivt_address - self.offset
        self.bdt.app_length = self.size + self.offset
        self.bdt.plugin = 1 if self.plugin else 0

    def info(self) -> str:
        """String representation of the IMX Boot Image."""
        self._update()
        # Print IVT
        msg = "#" * 60 + "\n"
        msg += "# IVT (Image Vector Table)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.ivt.info()
        # Print DBI
        msg += "#" * 60 + "\n"
        msg += "# BDI (Boot Data Info)\n"
        msg += "#" * 60 + "\n\n"
        msg += self.bdt.info()
        # Print DCD
        if self.dcd:
            msg += "#" * 60 + "\n"
            msg += "# DCD (Device Config Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.dcd.info()
        # Print CSF
        if self.csf.enabled:
            msg += "#" * 60 + "\n"
            msg += "# CSF (Code Signing Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.csf.info()
        return msg

    def add_image(
        self, data: bytes, img_type: EnumAppType = EnumAppType.APP, address: int = 0
    ) -> None:
        """Add specific image into the main boot image.

        :param data: Raw data of img
        :param img_type: Type of img
        :param address: address in RAM
        :raises Exception: raised when data type is unknown
        """
        if img_type == EnumAppType.APP:
            self.app.data = data
            if address != 0:
                self.address = address
        else:
            raise SPSDKError("Unknown data type !")

    def export(self) -> bytes:
        """Export Image as bytes array.

        :return: bytes
        """
        self._update()
        data = self.ivt.export()
        data += self.bdt.export()
        if self.dcd:
            data += self.dcd.export()
        data += self.app.export()
        data += self.csf.export()
        return data

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> BootImgBase:
        """Parse image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step
        :param size: parsing size
        :raises SPSDKError: Raised when the value type is incorrect
        :raises SPSDKError: Raised when there is not an i.MX Boot Image
        :return: BootImg2 object
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f"Not correct value type: {type(stream)}!")

        header = Header()
        start_index = stream.tell()
        last_index = stream.seek(0, SEEK_END)
        stream.seek(start_index)

        if size:
            last_index = min(start_index + size, last_index)

        imx_image = False
        while start_index < (last_index - Header.SIZE):
            header = Header.parse(read_raw_data(stream, Header.SIZE, no_seek=True))
            if (
                header.tag == SegTag.IVT2
                or header.length == SegIVT2.SIZE
                or header.param in (0x40, 0x41, 0x42, 0x43)
            ):
                imx_image = True
                break

            start_index = stream.seek(step, SEEK_CUR)

        if not imx_image:
            raise SPSDKError("Not an i.MX Boot Image!")

        obj = cls(version=header.param)
        img_size = last_index - start_index
        if start_index > 0:
            obj.offset = start_index

        # Parse IVT
        obj.ivt = SegIVT2.parse(read_raw_segment(stream, SegTag.IVT2))
        # Parse BDT
        obj.bdt = SegBDT.parse(read_raw_data(stream, SegBDT.SIZE))
        obj.offset = obj.ivt.ivt_address - obj.bdt.app_start
        obj.address = obj.bdt.app_start
        obj.plugin = bool(obj.bdt.plugin)
        # Parse DCD
        if obj.ivt.dcd_address:
            obj.dcd = SegDCD.parse(read_raw_segment(stream, SegTag.DCD))
            obj.dcd.padding = (obj.ivt.app_address - obj.ivt.dcd_address) - obj.dcd.size
        # Parse APP
        app_start = start_index + (obj.ivt.app_address - obj.ivt.ivt_address)
        app_size = (
            obj.ivt.csf_address - obj.ivt.app_address
            if obj.ivt.csf_address
            else obj.bdt.app_length - (obj.bdt.app_start - obj.ivt.app_address)
        )
        app_size = img_size - app_start if app_size > (img_size - app_start) else app_size
        obj.app.data = read_raw_data(stream, app_size, app_start)
        obj.app.padding = 0
        # Parse CSF
        # Finalize the code below
        # if obj.ivt.csf_address:
        #    obj.csf = SegCSF.parse(buffer)
        #    obj.csf.padding = obj.bdt.length - ((obj.ivt.csf_address - obj.ivt.ivt_address) + obj.csf.size)

        return obj


########################################################################################################################
# Boot Image V3a: i.MX8QXP-A0
########################################################################################################################


class BootImg3a(BootImgBase):
    """i.MX Boot Image v3a."""

    IMG_TYPE_CSF = 0x01
    IMG_TYPE_SCD = 0x02
    IMG_TYPE_EXEC = 0x03
    IMG_TYPE_DATA = 0x04

    SCFW_FLAGS_APP = 0x01355FC4
    SCFW_FLAGS_M4_0 = 0x4A5162
    SCFW_FLAGS_M4_1 = 0x4F52A3
    SCFW_FLAGS_SCFW = 0x1

    INITIAL_LOAD_ADDR_SCU_ROM = 0x2000E000
    INITIAL_LOAD_ADDR_AP_ROM = 0x00110000
    INITIAL_LOAD_ADDR_FLEXSPI = 0x08000000

    # The value of CSF segment size
    CSF_SIZE = 0x2000
    # The align value of APP segment
    IMG_AUTO_ALIGN = 0x10
    SECTOR_SIZE = 0x200
    APP_ALIGN = 0x1200
    # The value of img head size
    #           offset | size
    HEAD_SIZE = {0x400: 0xC400, 0x1000: 0x1400}

    PADDING_VAL = 0x00

    COUNT_OF_CONTAINERS = 2

    @property
    def plg(self) -> bool:
        """PLG."""
        return self._plg

    @plg.setter
    def plg(self, value: bool) -> None:
        assert isinstance(value, bool)
        self._plg = value

    @property
    def ivt(self) -> List[SegIVT3a]:
        """IVT."""
        return self._ivt

    @ivt.setter
    def ivt(self, value: List) -> None:
        assert isinstance(value, list) and isinstance(value[0], SegIVT3a)
        self._ivt = value

    @property
    def bdt(self) -> List[SegBDS3a]:
        """BDT."""
        return self._bdt

    @bdt.setter
    def bdt(self, value: List) -> None:
        assert isinstance(value, list) and isinstance(value[0], SegBDS3a)
        self._bdt = value

    @property
    def csf(self) -> SegCSF:
        """CSF."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        assert isinstance(value, SegCSF)
        self._csf = value

    def __init__(self, address: int = 0, offset: int = 0x400, version: int = 0x43) -> None:
        """Initialize boot image object.

        :param address: The start address of img in target memory
        :param offset: The IVT offset
        :param version: The version of boot img format
        """
        super().__init__(address, offset)
        self._ivt = [SegIVT3a(version), SegIVT3a(version)]
        self._ivt[0].next = self._ivt[0].size
        self._ivt[0].version = 0x01
        self._ivt[1].version = 0x01
        self._bdt = [SegBDS3a(), SegBDS3a()]
        self.app = [
            [SegAPP() for _ in range(SegBDS3a.IMAGES_MAX_COUNT)],
            [SegAPP() for _ in range(SegBDS3a.IMAGES_MAX_COUNT)],
        ]
        self._dcd = SegDCD()
        self._csf = SegCSF()
        self._plg = False
        if not isinstance(self.address, list):
            self.address = [self.INITIAL_LOAD_ADDR_SCU_ROM, self.INITIAL_LOAD_ADDR_AP_ROM]  # type: ignore
        self._sdc_address = 0

    @staticmethod
    def _compute_padding(size: int, sector_size: int) -> int:
        return (size // sector_size + (size % sector_size > 0)) * sector_size - size

    def _update(self) -> None:
        # Set zero padding for IVT and BDT sections
        for container in range(self.COUNT_OF_CONTAINERS):
            self.ivt[container].padding = 0
            self.bdt[container].padding = 0

            # Set IVT section
            self.ivt[container].ivt_address = (
                self.address[container]  # type: ignore
                + self.offset  # type: ignore
                + container * self.ivt[container].size  # type: ignore
            )
            self.ivt[container].bdt_address = (
                self.ivt[container].ivt_address
                + self.ivt[container].space * (self.COUNT_OF_CONTAINERS - container)
                + container * self.bdt[container].size
            )

            if container == 0:
                if self.dcd:
                    self.ivt[container].dcd_address = (
                        self.ivt[container].bdt_address + self.bdt[container].space * 2
                    )
                    if self.csf.enabled:
                        self.ivt[container].csf_address = (
                            self.ivt[container].dcd_address + self.dcd.space
                        )
                    else:
                        self.ivt[container].csf_address = 0
                else:
                    self.ivt[container].dcd_address = 0
                    if self.csf.enabled:
                        self.ivt[container].csf_address = (
                            self.ivt[container].bdt_address + self.bdt[container].space * 2
                        )
                    else:
                        self.ivt[container].csf_address = 0
            else:
                self.ivt[container].dcd_address = 0
                self.ivt[container].csf_address = 0

            self.app[container][0].padding = self._compute_padding(
                self.bdt[container].images[0].image_size, self.SECTOR_SIZE
            )
            if self.bdt[container].images_count != 0:
                self.bdt[container].boot_data_size = self.bdt[container].size
                if container == 0:
                    self.bdt[container].images[0].image_source = self.APP_ALIGN
                else:
                    last_image_index = self.bdt[container - 1].images_count - 1
                    last_image_address = (
                        self.bdt[container - 1].images[last_image_index].image_source
                    )
                    self.bdt[container].images[0].image_source = (
                        last_image_address + self.app[container - 1][last_image_index].space
                    )
            for i in range(self.bdt[container].images_count - 1):
                self.bdt[container].images[i + 1].image_source = (
                    self.bdt[container].images[i].image_source + self.app[container][i].space
                )
                self.app[container][i + 1].padding = self._compute_padding(
                    self.bdt[container].images[i + 1].image_size, self.SECTOR_SIZE
                )
            if container == self.COUNT_OF_CONTAINERS - 1:
                self.app[container][self.bdt[container].images_count - 1].padding = 0
                # Set BDT section

    def info(self) -> str:
        """String representation of the i.MX Boot Image v3a."""
        self._update()
        # Print IVT
        msg = "#" * 60 + "\n"
        msg += "# IVT (Image Vector Table)\n"
        msg += "#" * 60 + "\n\n"
        for index, ivt in enumerate(self.ivt):
            msg += "-" * 60 + "\n"
            msg += f"- IVT[{index}]\n"
            msg += "-" * 60 + "\n\n"
            msg += ivt.info()
        # Print BDI
        msg += "#" * 60 + "\n"
        msg += "# BDI (Boot Data Info)\n"
        msg += "#" * 60 + "\n\n"
        for index, bdi in enumerate(self.bdt):
            msg += "-" * 60 + "\n"
            msg += f"- BDI[{index}]\n"
            msg += "-" * 60 + "\n\n"
            msg += bdi.info()
        # Print DCD
        if self.dcd:
            msg += "#" * 60 + "\n"
            msg += "# DCD (Device Config Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.dcd.info()
        # Print CSF
        if self.csf.enabled:
            msg += "#" * 60 + "\n"
            msg += "# CSF (Code Signing Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.csf.info()
        return msg

    def add_image(
        self, data: bytes, img_type: EnumAppType = EnumAppType.APP, address: int = 0
    ) -> None:
        """Add specific image into the main boot image.

        :param data: Raw data of image
        :param img_type: Type of image
        :param address: address in RAM
        :raises Exception: raised when data type is unknown
        """
        if img_type == EnumAppType.APP:
            image_index = self.bdt[1].images_count
            self.bdt[1].images[image_index].image_destination = address
            self.bdt[1].images[image_index].image_entry = address
            self.bdt[1].images[image_index].image_size = len(data)
            self.bdt[1].images[image_index].rom_flags = 0
            self.bdt[1].images[image_index].hab_flags = self.IMG_TYPE_EXEC
            self.bdt[1].images[image_index].scfw_flags = self.SCFW_FLAGS_APP
            self.bdt[1].images_count += 1

            self.app[1][image_index].data = data
            self.app[1][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)

        elif img_type in (EnumAppType.M4_0, EnumAppType.M4_1):
            image_index = self.bdt[0].images_count
            self.bdt[0].images[image_index].image_destination = address
            self.bdt[0].images[image_index].image_entry = address
            self.bdt[0].images[image_index].image_size = len(data)
            self.bdt[0].images[image_index].rom_flags = 0
            self.bdt[0].images[image_index].hab_flags = self.IMG_TYPE_EXEC
            self.bdt[0].images[image_index].scfw_flags = (
                self.SCFW_FLAGS_M4_0 if img_type == EnumAppType.M4_0 else self.SCFW_FLAGS_M4_1
            )
            self.bdt[0].images_count += 1

            self.app[0][image_index].data = data
            self.app[0][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)

        elif img_type == EnumAppType.SCFW:
            image_index = self.bdt[0].images_count
            self.bdt[0].images[image_index].image_destination = 0x1FFE0000
            self.bdt[0].images[image_index].image_entry = 0x1FFE0000
            self.bdt[0].images[image_index].image_size = len(data)
            self.bdt[0].images[image_index].rom_flags = 0
            self.bdt[0].images[image_index].hab_flags = self.IMG_TYPE_EXEC
            self.bdt[0].images[image_index].scfw_flags = self.SCFW_FLAGS_SCFW
            self.bdt[0].images_count += 1

            self.app[0][image_index].data = data
            self.app[0][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)
            self._sdc_address = (
                self.bdt[0].images[image_index].image_destination
                + len(data)
                + self._compute_padding(len(data), self.IMG_AUTO_ALIGN)
            )

        elif img_type == EnumAppType.SCD:
            if self._sdc_address == 0:
                raise SPSDKError("SCFW have to be define before SCD!")
            image_index = self.bdt[0].images_count
            self.bdt[0].images[image_index].image_destination = self._sdc_address
            self.bdt[0].images[image_index].image_entry = 0
            self.bdt[0].images[image_index].image_size = len(data)
            self.bdt[0].images[image_index].rom_flags = 0
            self.bdt[0].images[image_index].hab_flags = self.IMG_TYPE_SCD
            self.bdt[0].images[image_index].scfw_flags = 0x1
            self.bdt[0].images_count += 1

            self.app[0][image_index].data = data
            self.app[0][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)

        else:
            raise SPSDKError("Unknown data type!")

    def export(self) -> bytes:
        """Export Image as binary blob."""
        self._update()
        data = bytes()
        data += self.ivt[0].export()
        data += self.ivt[1].export()
        data += self.bdt[0].export()
        data += self.bdt[1].export()
        if self.dcd:
            data += self.dcd.export()
        data += self.csf.export()
        data += bytes(
            [self.PADDING_VAL] * self._compute_padding(len(data), self.APP_ALIGN - self.offset)
        )

        for container in range(self.COUNT_OF_CONTAINERS):
            for image in range(self.bdt[container].images_count):
                data += self.app[container][image].export()

        return data

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> BootImgBase:
        """Parse image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step
        :param size: parsing size
        :raises SPSDKError: Raised when the values type is incorrect
        :raises SPSDKError: Raised when there is not an i.MX Boot Image
        :return: BootImg3a object
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f"Not correct value type: {type(stream)}!")

        header = Header()
        start_index = stream.tell()
        last_index = stream.seek(0, SEEK_END)
        stream.seek(start_index)

        if size:
            last_index = min(start_index + size, last_index)

        imx_image = False
        while start_index < (last_index - Header.SIZE):
            header = Header.parse(read_raw_data(stream, Header.SIZE, no_seek=True))
            if (
                header.tag == SegTag.IVT3
                or header.length == SegIVT3a.SIZE
                or header.param in (0x43,)
            ):
                imx_image = True
                break

            start_index = stream.seek(step, SEEK_CUR)

        if not imx_image:
            raise SPSDKError("Not an i.MX Boot Image!")

        obj = cls(version=header.param)
        if start_index > 0:
            obj.offset = start_index
        # Parse IVT
        obj.ivt[0] = SegIVT3a.parse(read_raw_segment(stream, SegTag.IVT3))
        obj.ivt[1] = SegIVT3a.parse(read_raw_segment(stream, SegTag.IVT3))
        # Parse BDT
        obj.bdt[0] = SegBDS3a.parse(read_raw_data(stream, SegBDS3a.SIZE))
        obj.bdt[1] = SegBDS3a.parse(read_raw_data(stream, SegBDS3a.SIZE))
        # Parse DCD
        if obj.ivt[0].dcd_address:
            stream.seek(start_index + (obj.ivt[0].dcd_address - obj.ivt[0].ivt_address), 0)
            obj.dcd = SegDCD.parse(read_raw_segment(stream, SegTag.DCD))
        # Parse CSF
        if obj.ivt[0].csf_address:
            stream.seek(start_index + (obj.ivt[0].csf_address - obj.ivt[0].ivt_address), 0)
            obj.csf = SegCSF.parse(read_raw_segment(stream, SegTag.CSF))
        # Parse IMAGES
        for container in range(obj.COUNT_OF_CONTAINERS):
            for i in range(obj.bdt[container].images_count):
                stream.seek(obj.bdt[container].images[i].image_source - obj.offset, 0)
                obj.app[container][i].data = read_raw_data(
                    stream, obj.bdt[container].images[i].image_size
                )

        return obj


########################################################################################################################
# Boot Image V3b: i.MX8QM-A0
########################################################################################################################


class BootImg3b(BootImgBase):
    """IMX Boot Image v3b."""

    IMG_TYPE_CSF = 0x01
    IMG_TYPE_SCD = 0x02
    IMG_TYPE_EXEC = 0x03
    IMG_TYPE_DATA = 0x04

    SCFW_FLAGS_A53 = 0x1354014
    SCFW_FLAGS_A72 = 0x1354065
    SCFW_FLAGS_M4_0 = 0x4A5162
    SCFW_FLAGS_M4_1 = 0x4F52A3
    SCFW_FLAGS_SCFW = 0x1

    INITIAL_LOAD_ADDR_SCU_ROM = 0x2000E000
    INITIAL_LOAD_ADDR_AP_ROM = 0x00110000
    INITIAL_LOAD_ADDR_FLEXSPI = 0x08000000

    # The value of CSF segment size
    CSF_SIZE = 0x2000
    # The align value for img
    IMG_AUTO_ALIGN = 0x10
    # The align value for sector
    SECTOR_SIZE = 0x200
    # The align value of APP segment
    APP_ALIGN = 0x1200

    PADDING_VAL = 0x00
    # The value of img head size
    #           offset | size
    HEAD_SIZE = {0x400: 0xC400, 0x1000: 0x1400}

    COUNT_OF_CONTAINERS = 2

    @property
    def plg(self) -> bool:
        """PLG."""
        return self._plg

    @plg.setter
    def plg(self, value: bool) -> None:
        assert isinstance(value, bool)
        self._plg = value

    @property
    def ivt(self) -> List[SegIVT3b]:
        """IVT."""
        return self._ivt

    @ivt.setter
    def ivt(self, value: List) -> None:
        assert isinstance(value, list)
        if len(value) != self.COUNT_OF_CONTAINERS:
            raise SPSDKError("Invalid value of IVT")
        assert isinstance(value[0], SegIVT3b)
        self._ivt = value

    @property
    def bdt(self) -> List[SegBDS3b]:
        """BDT."""
        return self._bdt

    @bdt.setter
    def bdt(self, value: List) -> None:
        assert isinstance(value, list)
        if len(value) != self.COUNT_OF_CONTAINERS:
            raise SPSDKError("Invalid value of BDT")
        assert isinstance(value[0], SegBDS3b)
        self._bdt = value

    @property
    def csf(self) -> SegCSF:
        """CSF."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        assert isinstance(value, SegCSF)
        self._csf = value

    def __init__(self, address: int = 0, offset: int = 0x400, version: int = 0x43) -> None:
        """Initialize boot image object.

        :param address: The start address of img in target memory
        :param offset: The IVT offset
        :param version: The version of boot img format
        """
        super().__init__(address, offset)
        self._ivt = [SegIVT3b(version), SegIVT3b(version)]
        self._bdt = [SegBDS3b(), SegBDS3b()]
        self.app = [
            [SegAPP() for _ in range(SegBDS3b.IMAGES_MAX_COUNT)],
            [SegAPP() for _ in range(SegBDS3b.IMAGES_MAX_COUNT)],
        ]
        self._dcd = SegDCD()
        self.scd = SegAPP()
        self._csf = SegCSF()
        self._plg = False
        self._scd_address = 0
        if not isinstance(self.address, int):
            self.address = [self.INITIAL_LOAD_ADDR_SCU_ROM, self.INITIAL_LOAD_ADDR_AP_ROM]  # type: ignore

    @staticmethod
    def _compute_padding(image_size: int, sector_size: int) -> int:
        return (
            (image_size // sector_size + (image_size % sector_size > 0)) * sector_size
        ) - image_size

    def _update(self) -> None:
        # Set zero padding for IVT and BDT sections
        for container in range(self.COUNT_OF_CONTAINERS):
            self.ivt[container].padding = 0
            self.bdt[container].padding = 0

            # Set IVT section
            self.ivt[container].ivt_address = (
                self.address[container]  # type: ignore
                + self.offset  # type: ignore
                + container * self.ivt[container].size  # type: ignore
            )
            self.ivt[container].bdt_address = (
                self.ivt[container].ivt_address
                + self.ivt[container].space * (2 - container)
                + container * self.bdt[container].size
            )
            if container == 0:
                if self.dcd:
                    self.ivt[container].dcd_address = (
                        self.ivt[container].bdt_address + self.bdt[container].space * 2
                    )
                    if self.csf.enabled:
                        self.ivt[container].csf_address = (
                            self.ivt[container].dcd_address + self.dcd.space
                        )
                    else:
                        self.ivt[container].csf_address = 0
                else:
                    self.ivt[container].dcd_address = 0
                    if self.csf.enabled:
                        self.ivt[container].csf_address = (
                            self.ivt[container].bdt_address + self.bdt[container].space * 2
                        )
                    else:
                        self.ivt[container].csf_address = 0
            else:
                self.ivt[container].dcd_address = 0
                self.ivt[container].csf_address = 0

            self.app[container][0].padding = self._compute_padding(
                self.bdt[container].images[0].image_size, self.SECTOR_SIZE
            )
            if self.bdt[container].images_count != 0:
                self.bdt[container].boot_data_size = self.bdt[container].size
                if container == 0:
                    self.bdt[container].images[0].image_source = self.APP_ALIGN
                else:
                    last_image_index = self.bdt[container - 1].images_count - 1
                    last_image_address = (
                        self.bdt[container - 1].images[last_image_index].image_source
                    )
                    self.bdt[container].images[0].image_source = (
                        last_image_address + self.app[container - 1][last_image_index].space
                    )
            next_image_address = 0
            for i in range(self.bdt[container].images_count - 1):
                self.bdt[container].images[i + 1].image_source = (
                    self.bdt[container].images[i].image_source + self.app[container][i].space
                )
                self.app[container][i + 1].padding = self._compute_padding(
                    self.bdt[container].images[i + 1].image_size, self.SECTOR_SIZE
                )
                next_image_address = (
                    self.bdt[container].images[i + 1].image_source
                    + self.app[container][i + 1].space
                )

            if container == 0:
                if self.bdt[container].scd.image_destination != 0:
                    self.bdt[container].scd.image_source = next_image_address
                    self.scd.padding = self._compute_padding(
                        self.bdt[0].scd.image_size, self.SECTOR_SIZE
                    )
                    next_image_address += self.scd.space
                    # Set BDT section

                if self.csf.enabled:
                    self.bdt[container].csf.image_source = next_image_address
                    self.csf.padding = self._compute_padding(
                        self.bdt[0].csf.image_size, self.SECTOR_SIZE
                    )
                    next_image_address += self.csf.space
                    # Set BDT section

    def info(self) -> str:
        """String representation of the IMX Boot Image v3b."""
        self._update()
        # Print IVT
        msg = "#" * 60 + "\n"
        msg += "# IVT (Image Vector Table)\n"
        msg += "#" * 60 + "\n\n"
        for index, ivt in enumerate(self.ivt):
            msg += "-" * 60 + "\n"
            msg += f"- IVT[{index}]\n"
            msg += "-" * 60 + "\n\n"
            msg += ivt.info()
        # Print BDI
        msg += "#" * 60 + "\n"
        msg += "# BDI (Boot Data Info)\n"
        msg += "#" * 60 + "\n\n"
        for index, bdi in enumerate(self.bdt):
            msg += "-" * 60 + "\n"
            msg += f"- BDI[{index}]\n"
            msg += "-" * 60 + "\n\n"
            msg += bdi.info()
        # Print DCD
        if self.dcd:
            msg += "#" * 60 + "\n"
            msg += "# DCD (Device Config Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.dcd.info()
        # Print CSF
        if self.csf.enabled:
            msg += "#" * 60 + "\n"
            msg += "# CSF (Code Signing Data)\n"
            msg += "#" * 60 + "\n\n"
            msg += self.csf.info()
        return msg

    def add_image(
        self, data: bytes, img_type: EnumAppType = EnumAppType.APP, address: int = 0
    ) -> None:
        """Add specific image into the main boot image.

        :param data: Raw data of image
        :param img_type: Type of image
        :param address: address in RAM
        :raises Exception: raised SCFW is not defined before SCD
        :raises Exception: raised when there is unknown image type
        """
        if img_type in (EnumAppType.APP, EnumAppType.A72):
            image_index = self.bdt[1].images_count
            self.app[1][image_index].data = data

            self.bdt[1].images[image_index].image_destination = address
            self.bdt[1].images[image_index].image_entry = address
            self.bdt[1].images[image_index].image_size = len(data)

            if img_type == EnumAppType.APP:
                self.bdt[1].images[image_index].flags = self.SCFW_FLAGS_A53
            elif img_type == EnumAppType.A72:
                self.bdt[1].images[image_index].flags = self.SCFW_FLAGS_A72

            self.app[1][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)
            self.bdt[1].images_count += 1

        elif img_type in (EnumAppType.M4_0, EnumAppType.M4_1):
            image_index = self.bdt[0].images_count
            self.app[0][image_index].data = data

            self.bdt[0].images[image_index].image_destination = address
            self.bdt[0].images[image_index].image_entry = address
            self.bdt[0].images[image_index].image_size = len(data)

            if img_type == EnumAppType.M4_0:
                self.bdt[0].images[image_index].flags = self.SCFW_FLAGS_M4_0
            elif img_type == EnumAppType.M4_1:
                self.bdt[0].images[image_index].flags = self.SCFW_FLAGS_M4_1

            self.app[0][image_index].padding = (
                (len(data) // self.SECTOR_SIZE + (len(data) % self.SECTOR_SIZE > 0))
                * self.SECTOR_SIZE
            ) - len(data)
            self.bdt[0].images_count += 1

        elif img_type == EnumAppType.SCFW:
            image_index = self.bdt[0].images_count
            self.bdt[0].images[image_index].image_destination = 0x30FE0000
            self.bdt[0].images[image_index].image_entry = 0x1FFE0000
            self.bdt[0].images[image_index].image_size = len(data)
            self.bdt[0].images[image_index].flags = self.SCFW_FLAGS_SCFW
            self._scd_address = (
                self.bdt[0].images[image_index].image_destination
                + len(data)
                + self._compute_padding(len(data), self.IMG_AUTO_ALIGN)
            )
            self.bdt[0].images_count += 1

            self.app[0][image_index].data = data
            self.app[0][image_index].padding = self._compute_padding(len(data), self.SECTOR_SIZE)

        elif img_type == EnumAppType.SCD:
            if self._scd_address == 0:
                raise SPSDKError("SCFW have to be define before SCD!")
            self.scd.data = data
            self.scd.padding = self._compute_padding(len(data), self.SECTOR_SIZE)
            self.bdt[0].scd.image_destination = self._scd_address
            self.bdt[0].scd.image_entry = 0
            self.bdt[0].scd.image_size = len(data)
            self.ivt[0].scd_address = self.bdt[0].scd.image_destination

        else:
            raise SPSDKError("Unknown image type!")

    def export(self) -> bytes:
        """Export."""
        self._update()
        # data = bytearray(self._offset)
        data = bytes()
        data += self.ivt[0].export()
        data += self.ivt[1].export()
        data += self.bdt[0].export()
        data += self.bdt[1].export()
        if self.dcd:
            data += self.dcd.export()
        data += bytes(
            [self.PADDING_VAL] * self._compute_padding(len(data), self.APP_ALIGN - self.offset)
        )

        for container in range(self.COUNT_OF_CONTAINERS):
            for i in range(self.bdt[container].images_count):
                data += self.app[container][i].export()

        if self.bdt[0].scd.image_source != 0:
            data += self.scd.export()

        if self.bdt[0].csf.image_source != 0:
            data += self.csf.export()

        return data

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> BootImgBase:
        """Parse image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step
        :param size: parsing size
        :raises SPSDKError: When the value is incorrect
        :raises SPSDKError: If there is not an i.MX Boot Image
        :return: BootImg3b object
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f"Not correct value type: {type(stream)}!")

        header = Header()
        start_index = stream.tell()
        last_index = stream.seek(0, SEEK_END)
        stream.seek(start_index)

        if size:
            last_index = min(start_index + size, last_index)

        imx_image = False
        while start_index < (last_index - Header.SIZE):
            header = Header.parse(read_raw_data(stream, Header.SIZE, no_seek=True))
            if (
                header.tag == SegTag.IVT2
                or header.length == SegIVT3b.SIZE
                or header.param in (0x43,)
            ):
                imx_image = True
                break

            start_index = stream.seek(step, SEEK_CUR)

        if not imx_image:
            raise SPSDKError("Not an i.MX Boot Image!")

        obj = cls(version=header.param)
        if start_index > 0:
            obj.offset = start_index
        # Parse IVT
        obj.ivt[0] = SegIVT3b.parse(read_raw_segment(stream, SegTag.IVT2))
        obj.ivt[1] = SegIVT3b.parse(read_raw_segment(stream, SegTag.IVT2))
        # Parse BDT
        obj.bdt[0] = SegBDS3b.parse(read_raw_data(stream, SegBDS3b.SIZE))
        obj.bdt[1] = SegBDS3b.parse(read_raw_data(stream, SegBDS3b.SIZE))
        # Parse DCD
        if obj.ivt[0].dcd_address:
            stream.seek(start_index + (obj.ivt[0].dcd_address - obj.ivt[0].ivt_address), 0)
            obj.dcd = SegDCD.parse(read_raw_segment(stream, SegTag.DCD))
        # Parse IMAGES
        for container in range(obj.COUNT_OF_CONTAINERS):
            for i in range(obj.bdt[container].images_count):
                stream.seek(obj.bdt[container].images[i].image_source - obj.offset, 0)
                obj.app[container][i].data = read_raw_data(
                    stream, obj.bdt[container].images[i].image_size
                )
        # Parse SCD
        if obj.bdt[0].scd.image_source != 0:
            stream.seek(obj.bdt[0].scd.image_source - obj.offset, 0)
            obj.scd.data = read_raw_data(stream, obj.bdt[0].scd.image_size)
        # Parse CSF
        if obj.bdt[0].csf.image_source != 0:
            stream.seek(obj.bdt[0].csf.image_source - obj.offset, 0)
            obj.csf = SegCSF.parse(read_raw_segment(stream, SegTag.CSF))

        return obj


########################################################################################################################
# Boot Image V4: i.MX8DM, i.MX8QM_B0, i.MX8QXP_B0
########################################################################################################################


class BootImg4(BootImgBase):
    """i.MX Boot Image v4."""

    def __init__(self, address: int = 0, offset: int = 0x400) -> None:
        """Initialize boot image object.

        :param address: The start address of image in target memory
        :param offset: The image offset
        """
        super().__init__(address, offset)
        self._dcd = SegDCD()
        self._cont1_header = SegBIC1()
        self._cont2_header = SegBIC1()

    def _update(self) -> None:
        pass

    def info(self) -> str:
        """String representation of the i.MX Boot Image v4."""
        self._update()
        msg = ""
        msg += "#" * 60 + "\n"
        msg += "# Boot Images Container 1\n"
        msg += "#" * 60 + "\n\n"
        msg += self._cont1_header.info()
        msg += "#" * 60 + "\n"
        msg += "# Boot Images Container 2\n"
        msg += "#" * 60 + "\n\n"
        msg += self._cont2_header.info()
        if self.dcd:
            if self.dcd.enabled:
                msg += "#" * 60 + "\n"
                msg += "# DCD (Device Config Data)\n"
                msg += "#" * 60 + "\n\n"
                msg += self.dcd.info()
        return msg

    def add_image(self, data: bytes, img_type: int, address: int) -> None:
        """Add image.

        :raises NotImplementedError: Not yet implemented
        """
        raise NotImplementedError("Not yet implemented.")

    def export(self) -> bytes:
        """Export."""
        self._update()
        data = bytes()
        data += self._cont1_header.export()
        data += self._cont2_header.export()
        # Complete Implementation
        return data

    @classmethod
    def parse(
        cls,
        stream: Union[bytes, bytearray, BufferedReader, BytesIO],
        step: int = 0x100,
        size: Optional[int] = None,
    ) -> BootImgBase:
        """Parse image from stream buffer or bytes array.

        :param stream: The stream buffer or bytes array
        :param step: Image searching step
        :param size: parsing size
        :return: BootImg4 object
        :raises SPSDKError: Raised when the value type is incorrect
        :raises SPSDKError: If there is not an i.MX Boot Image
        """
        if isinstance(stream, (bytes, bytearray)):
            stream = BytesIO(stream)

        if not isinstance(stream, (BufferedReader, BytesIO)):
            raise SPSDKError(f" Not correct value type: '{type(stream)}' !")

        start_index = stream.tell()
        last_index = stream.seek(0, SEEK_END)
        stream.seek(start_index)

        if size:
            last_index = min(start_index + size, last_index)

        imx_image = False
        while start_index < (last_index - Header.SIZE):
            header = Header.parse(read_raw_data(stream, Header2.SIZE, no_seek=True))
            if header.tag == SegTag.BIC1:
                imx_image = True
                break

            start_index = stream.seek(step, SEEK_CUR)

        if not imx_image:
            raise SPSDKError(" Not an i.MX Boot Image !")

        obj = cls()
        if start_index > 0:
            obj.offset = start_index

        # Parse Containers
        obj._cont1_header = SegBIC1.parse(read_raw_data(stream, 0x400))
        obj._cont2_header = SegBIC1.parse(read_raw_data(stream, 0x400))
        # Complete Implementation
        return obj


########################################################################################################################
# i.MX Kernel Image Classes
########################################################################################################################


class KernelImg:
    """IMX Kernel Image."""

    IMAGE_MIN_SIZE = 0x1000

    @property
    def address(self) -> int:
        """Address."""
        return self._ivt.app_address

    @address.setter
    def address(self, value: int) -> None:
        self._ivt.app_address = value

    @property
    def version(self) -> int:
        """Version."""
        return self._ivt.version

    @version.setter
    def version(self, value: int) -> None:
        self._ivt.version = value

    @property
    def app(self) -> Optional[bytes]:
        """APP."""
        return self._app.data

    @app.setter
    def app(self, value: Union[bytes, bytearray]) -> None:
        assert isinstance(value, (bytes, bytearray))
        self._app.data = value

    @property
    def csf(self) -> SegCSF:
        """CSF."""
        return self._csf

    @csf.setter
    def csf(self, value: SegCSF) -> None:
        assert isinstance(value, SegCSF)
        self._csf = value

    def __init__(
        self,
        address: int = 0,
        app: Optional[bytes] = None,
        csf: Optional[Union[SegCSF, Any]] = None,
        version: int = 0x41,
    ) -> None:
        """Initialize the IMX Kernel Image."""
        self._ivt = SegIVT2(version)
        self._ivt.app_address = address
        self._app = SegAPP(app)
        self._csf = SegCSF() if csf is None else csf

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:
        return ""

    def _update(self) -> None:
        pass

    def info(self) -> None:
        """String representation of the IMX Kernel Image."""

    def export(self) -> bytes:
        """Export."""
        self._update()
        data = self._app.export()
        data += self._ivt.export()
        data += self._csf.export()
        return data

    @classmethod
    def parse(cls, data: Union[str, bytes]) -> None:
        """Parse."""
        assert isinstance(data, (bytes, str))
        if not len(data) > cls.IMAGE_MIN_SIZE:
            raise SPSDKError("Invalid length of data to be parsed")


########################################################################################################################
# i.MX Image Public Methods
########################################################################################################################


def parse(
    stream: Union[bytes, bytearray, BufferedReader, BytesIO],
    step: int = 0x100,
    size: Optional[int] = None,
) -> BootImgBase:
    """Common parser for all versions of i.MX boot images.

    :param stream: stream buffer to image
    :param step: Image searching step
    :param size: parsing size
    :return: the object of boot image
    :raises SPSDKError: Raised when the format of string is incorrect
    :raises SPSDKError: When not i.MX Boot Image is passed
    """
    if isinstance(stream, (bytes, bytearray)):
        stream = BytesIO(stream)

    if not isinstance(stream, (BufferedReader, BytesIO)):
        raise SPSDKError(f"Not correct value type: '{type(stream)}' !")

    # calculate stream size
    start_index = stream.tell()
    last_index = stream.seek(0, SEEK_END)
    stream.seek(start_index)

    if size:
        last_index = min(start_index + size, last_index)

    while start_index < (last_index - Header.SIZE):
        raw = read_raw_data(stream, Header.SIZE, no_seek=True)

        if (
            raw[0] == SegTag.IVT2
            and ((raw[1] << 8) | raw[2]) == SegIVT2.SIZE
            and raw[3] in (0x40, 0x41, 0x42)
        ):
            return BootImg2.parse(stream)

        if (
            raw[0] == SegTag.IVT2
            and ((raw[1] << 8) | raw[2]) == SegIVT3b.SIZE
            and raw[3] in (0x43,)
        ):
            return BootImg3b.parse(stream)

        if (
            raw[0] == SegTag.IVT3
            and ((raw[1] << 8) | raw[2]) == SegIVT3a.SIZE
            and raw[3] in (0x43,)
        ):
            return BootImg3a.parse(stream)

        if raw[3] == SegTag.BIC1:
            return BootImg4.parse(stream)

        start_index = stream.seek(step, SEEK_CUR)

    raise SPSDKError(" Not an i.MX Boot Image !")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""NXP USB Device Scanner API."""


import logging
from typing import List, Optional

from libusbsio import LIBUSBSIO_Exception, usbsio
from serial.tools.list_ports import comports

from spsdk import SPSDKError
from spsdk.mboot.interfaces.sdio import scan_sdio as mb_scan_sdio
from spsdk.mboot.interfaces.uart import scan_uart as mb_scan_uart
from spsdk.sdp import SDP
from spsdk.sdp.exceptions import SdpConnectionError
from spsdk.sdp.interfaces.uart import Uart as SDP_Uart

from .devicedescription import (
    SDIODeviceDescription,
    SIODeviceDescription,
    UartDeviceDescription,
    USBDeviceDescription,
    convert_usb_path,
    get_usb_device_name,
)

NXP_USB_DEVICE_VIDS = [
    0x1FC9,
    0x15A2,
]

NXP_SDIO_DEVICE_PATHS = [
    "/dev/mcu-sdio",
]

logger = logging.getLogger(__name__)


def search_nxp_sdio_devices() -> List[SDIODeviceDescription]:
    """Searches all NXP SDIO devices based on their device path.

    :return: list of SDIODeviceDescription corresponding to NXP devices
    """
    nxp_sdio_devices = []

    search_path = NXP_SDIO_DEVICE_PATHS

    for path in search_path:
        sdio_device = mb_scan_sdio(device_path=path)
        if len(sdio_device) > 0:
            sdio_dev = SDIODeviceDescription(
                vid=sdio_device[0].vid, pid=sdio_device[0].pid, path=sdio_device[0].path
            )
            nxp_sdio_devices.append(sdio_dev)
            continue

    return nxp_sdio_devices


def search_nxp_usb_devices(extend_vid_list: Optional[list] = None) -> List[USBDeviceDescription]:
    """Searches all NXP USB devices based on their Vendor ID.

    :extend_vid_list: list of VIDs, to extend the default NXP VID list (int)
    :return: list of USBDeviceDescription corresponding to NXP devices
    """
    libusbsio_logger = logging.getLogger("libusbsio")
    sio = usbsio(loglevel=libusbsio_logger.getEffectiveLevel())
    all_usb_devices = sio.HIDAPI_Enumerate()
    nxp_usb_devices = []

    search_vids = NXP_USB_DEVICE_VIDS

    if extend_vid_list is not None:
        search_vids = list(set(search_vids) | set(extend_vid_list))

    for usb_device in all_usb_devices:
        for nxp_vid in search_vids:
            if nxp_vid == usb_device["vendor_id"]:
                # We found our device, let's create container for it
                vid = usb_device["vendor_id"]
                pid = usb_device["product_id"]
                path = convert_usb_path(usb_device["path"])
                product_string = usb_device["product_string"]
                manufacturer_string = usb_device["manufacturer_string"]
                name = ", ".join(get_usb_device_name(vid, pid, None))
                serial = usb_device["serial_number"]
                usb_dev = USBDeviceDescription(
                    vid,
                    pid,
                    path,
                    product_string,
                    manufacturer_string,
                    name,
                    serial,
                    usb_device["path"],
                )

                nxp_usb_devices.append(usb_dev)
                break

    return nxp_usb_devices


def search_nxp_uart_devices() -> List[UartDeviceDescription]:
    """Returns a list of all NXP devices connected via UART.

    :retval: list of UartDeviceDescription devices from devicedescription module
    """
    retval = []

    # Get all available COM ports on target PC
    ports = [port.device for port in comports()]

    # Iterate over every com port we have and check, whether mboot or sdp responds
    for port in ports:
        if mb_scan_uart(port=port, timeout=50):
            uart_dev = UartDeviceDescription(name=port, dev_type="mboot device")
            retval.append(uart_dev)
            continue

        # Seems the port is not mboot, let's try SDP protocol
        # The SDP protocol is on uart interface, so opening just the port is not
        # sufficient, to say, that the interface is SDP compared to mboot, where
        # ping command must be sent.
        # So we create an SDP interface and try to read the status code. If
        # we get a response, we are connected to an SDP device.
        try:
            sdp_com = SDP(SDP_Uart(port=port, timeout=50))
            if sdp_com.read_status() is not None:
                uart_dev = UartDeviceDescription(name=port, dev_type="SDP device")
                retval.append(uart_dev)
        except SdpConnectionError as e:
            logger.debug(
                f"Exception {type(e).__name__} occurred while reading status via SDP. \
Arguments: {e.args}"
            )

    return retval


# This function has been left for potential future uses. At the moment it's
# not clear, how do we identify different devices with SDP protocol, as
# in the company, there are so many different MCU's from NXP and Freescale
# with different MCU identification options, if any...
# def parse_sim_sdid(sim_sdid: int) -> Dict:
#     """Converts the content of SIM_SDID register into string.

#     :sim_sdid: the value of SIM_SDID register
#     :retval: {"family_id": int, "subfamily_id": int, "series_id": int, "rev_id": int}
#     """
#     retval = {}
#     retval["family_id"] = (sim_sdid >> 28) & 0xF
#     retval["subfamily_id"] = (sim_sdid >> 24) & 0xF
#     retval["series_id"] = (sim_sdid >> 20) & 0xF
#     retval["rev_id"] = (sim_sdid >> 12) & 0xF
#     return retval


def search_libusbsio_devices() -> List[SIODeviceDescription]:
    """Returns a list of all LIBUSBSIO devices.

    :retval: list of UartDeviceDescription devices from devicedescription module
    :raises SPSDKError: In any case of LIBUSBSIO problems.
    """
    retval = []
    try:
        libusbsio_logger = logging.getLogger("libusbsio")
        sio = usbsio(loglevel=libusbsio_logger.getEffectiveLevel())
        for i in range(sio.GetNumPorts()):
            info = sio.GetDeviceInfo(i)
            if not info:
                raise SPSDKError("Cannot retrieve the Device Information.")
            retval.append(SIODeviceDescription(info))
    except (LIBUSBSIO_Exception, SPSDKError) as exc:
        raise SPSDKError(f"LIBUSBSIO search devices fails: [{str(exc)}]") from exc

    return retval



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Header."""

from struct import calcsize, pack, unpack_from
from typing import Optional

from spsdk import SPSDKError
from spsdk.exceptions import SPSDKParsingError
from spsdk.utils.easy_enum import Enum

########################################################################################################################
# Enums
########################################################################################################################


class SegTag(Enum):
    """Segments Tag."""

    XMCD = (0xC0, "XMCD", "External Memory Configuration Data")
    DCD = (0xD2, "DCD", "Device Configuration Data")
    CSF = (0xD4, "CSF", "Command Sequence File Data")
    # i.MX6, i.MX7, i.MX8M
    IVT2 = (0xD1, "IVT2", "Image Vector Table (Version 2)")
    CRT = (0xD7, "CRT", "Certificate")
    SIG = (0xD8, "SIG", "Signature")
    EVT = (0xDB, "EVT", "Event")
    RVT = (0xDD, "RVT", "ROM Vector Table")
    WRP = (0x81, "WRP", "Wrapped Key")
    MAC = (0xAC, "MAC", "Message Authentication Code")
    # i.MX8QXP_A0, i.MX8QM_A0
    IVT3 = (0xDE, "IVT3", "Image Vector Table (Version 3)")
    # i.MX8QXP_B0, i.MX8QM_B0
    BIC1 = (0x87, "BIC1", "Boot Images Container")
    SIGB = (0x90, "SIGB", "Signature block")


class CmdTag(Enum):
    """CSF/DCD Command Tag."""

    SET = (0xB1, "SET", "Set")
    INS_KEY = (0xBE, "INS_KEY", "Install Key")
    AUT_DAT = (0xCA, "AUT_DAT", "Authenticate Data")
    WRT_DAT = (0xCC, "WRT_DAT", "Write Data")
    CHK_DAT = (0xCF, "CHK_DAT", "Check Data")
    NOP = (0xC0, "NOP", "No Operation (NOP)")
    INIT = (0xB4, "INIT", "Initialize")
    UNLK = (0xB2, "UNLK", "Unlock")


########################################################################################################################
# Classes
########################################################################################################################


class Header:
    """Header element type."""

    FORMAT = ">BHB"
    SIZE = calcsize(FORMAT)

    @property
    def size(self) -> int:
        """Header size in bytes."""
        return self.SIZE

    def __init__(self, tag: int = 0, param: int = 0, length: Optional[int] = None) -> None:
        """Constructor.

        :param tag: section tag
        :param param: TODO
        :param length: length of the segment or command; if not specified, size of the header is used
        :raises SPSDKError: If invalid length
        """
        self._tag = tag
        self.param: int = param
        self.length: int = self.SIZE if length is None else length
        if self.SIZE > self.length or self.length >= 65536:
            raise SPSDKError("Invalid length")

    @property
    def tag(self) -> int:
        """:return: section tag: command tag or segment tag, ..."""
        return self._tag

    @property
    def tag_name(self) -> str:
        """Returns the header's tag name."""
        return SegTag.name(self.tag)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tag_name}, {self.param}, {self.length})"

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__} <TAG:{self.tag_name} 0x{self.tag:02X}, "
            f"PARAM:0x{self.param:02X}, LEN:{self.length}B>"
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and vars(other) == vars(self)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def info(self) -> str:
        """Text representation of the header."""
        return str(self)

    def export(self) -> bytes:
        """Binary representation of the header."""
        return pack(self.FORMAT, self.tag, self.length, self.param)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0, required_tag: Optional[int] = None) -> "Header":
        """Parse header.

        :param data: Raw data as bytes or bytearray
        :param offset: Offset of input data
        :param required_tag: Check header TAG if specified value or ignore if is None
        :return: Header object
        :raises SPSDKParsingError: if required header tag does not match
        """
        tag, length, param = unpack_from(cls.FORMAT, data, offset)
        if required_tag is not None and tag != required_tag:
            raise SPSDKParsingError(
                f" Invalid header tag: '0x{tag:02X}' expected '0x{required_tag:02X}' "
            )

        return cls(tag, param, length)


class CmdHeader(Header):
    """Command header."""

    def __init__(self, tag: CmdTag, param: int = 0, length: Optional[int] = None) -> None:
        """Constructor.

        :param tag: command tag
        :param param: TODO
        :param length: of the command binary section, in bytes
        :raises SPSDKError: If invalid command tag
        """
        super().__init__(tag, param, length)
        if tag not in CmdTag.tags():
            raise SPSDKError("Invalid command tag")

    @property
    def tag(self) -> CmdTag:
        """Command tag."""
        return CmdTag.from_int(self._tag)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0, required_tag: Optional[int] = None) -> Header:
        """Create Header from binary data.

        :param data: binary data to convert into header
        :param offset: to start reading binary data
        :param required_tag: CmdTag, None if not required
        :return: parsed instance
        :raises SPSDKParsingError: If required header tag does not match
        :raises SPSDKError: If invalid tag
        """
        if required_tag is not None:
            if required_tag not in CmdTag.tags():
                raise SPSDKError("Invalid tag")
        return super(CmdHeader, cls).parse(data, offset, required_tag)


class Header2(Header):
    """Header element type."""

    FORMAT = "<BHB"

    def export(self) -> bytes:
        """Binary representation of the header."""
        return pack(self.FORMAT, self.param, self.length, self.tag)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0, required_tag: Optional[int] = None) -> "Header":
        """Parse header.

        :param data: Raw data as bytes or bytearray
        :param offset: Offset of input data
        :param required_tag: Check header TAG if specified value or ignore if is None
        :raises SPSDKParsingError: Raises an error if required tag is empty or not valid
        :return: Header2 object
        """
        param, length, tag = unpack_from(cls.FORMAT, data, offset)
        if required_tag is not None and tag != required_tag:
            raise SPSDKParsingError(
                f" Invalid header tag: '0x{tag:02X}' expected '0x{required_tag:02X}' "
            )

        return cls(tag, param, length)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""CLI helper for Click."""

import logging
from gettext import gettext
from typing import Any, Callable, List, Optional, Sequence, Tuple, TypeVar, Union

import click
from click_command_tree import _build_command_tree, _CommandWrapper

from spsdk import __version__ as spsdk_version

FC = TypeVar("FC", bound=Union[Callable[..., Any], click.Command])

_uart_option = click.option(
    "-p",
    "--port",
    metavar="COM[,speed]",
    help="""Serial port configuration. Default baud rate is 57600.
    Use 'nxpdevscan' utility to list devices on serial port.""",
)

_sdp_uart_option = click.option(
    "-p",
    "--port",
    metavar="COM[,speed]",
    help="""Serial port configuration. Default baud rate is 115200.
    Use 'nxpdevscan' utility to list devices on serial port.""",
)

_usb_option = click.option(
    "-u",
    "--usb",
    metavar="VID:PID|USB_PATH|DEV_NAME",
    help="""USB device identifier.

    \b
    Following formats are supported: <vid>, <vid:pid> or <vid,pid>, device/instance path, device name.
    <vid>: hex or dec string; e.g. 0x0AB12, 43794.
    <vid/pid>: hex or dec string; e.g. 0x0AB12:0x123, 1:3451.
    Use 'nxpdevscan' utility to list connected device names.
""",
)

_sdio_option = click.option(
    "-sd",
    "--sdio",
    metavar="SDIO_PATH|DEV_NAME",
    help="""SDIO device identifier.

    \b
    Following formats are supported: device/instance path, device name.
    device/instance path: device string; e.g. /dev/mcu-sdio.
    Use 'nxpdevscan' utility to list connected device names.
""",
)

_lpcusbsio_option = click.option(
    "-l",
    "--lpcusbsio",
    metavar="[usb,VID:PID|USB_PATH|SER_NUM,]spi|i2c",
    help="""USB-SIO bridge interface.

    Optional USB device filtering formats:
    [usb,vid:pid|usb_path|serial_number]

    Following serial interfaces are supported:

    \b
    spi[index][,port,pin,speed_kHz,polarity,phase]
     - index ... optional index of SPI peripheral. Example: "spi1" (default=0)
     - port ... bridge GPIO port used as SPI SSEL(default=0)
     - pin  ... bridge GPIO pin used as SPI SSEL
        default SSEL is set to 0.15 which works
        for the LPCLink2 bridge. The MCULink OB
        bridge ignores the SSEL value anyway.(default=15)
     - speed_kHz ... SPI clock in kHz (default 1000)
     - polarity ... SPI CPOL option (default=1)
     - phase ... SPI CPHA option (default=1)

    \b
    i2c[index][,address,speed_kHz]
     - index ... optional index of I2C peripheral. Example: "i2c1" (default=0)
     - address ... I2C device address (default 0x10)
     - speed_kHz ... I2C clock in kHz (default 100)
""",
)

_buspal_option = click.option(
    "-b",
    "--buspal",
    metavar="spi[,speed,polarity,phase,lsb|msb] | i2c[,address,speed]",
    help="buspal settings",
)

_json_option = click.option(
    "-j",
    "--json",
    "use_json",
    is_flag=True,
    help="Use JSON output",
)


def _timeout_option(use_long_option: bool, timeout: int) -> Callable[[FC], FC]:
    """Get the timeout option.

    :param use_long_option: Use long version only
    :param timeout: Default timeout in miliseconds

    :return: click decorator
    """
    options = [] if use_long_option else ["-t"]
    options.append("--timeout")
    return click.option(
        *options,
        metavar="<ms>",
        help=f"""Sets timeout when waiting on data over a serial line. The default is {timeout} milliseconds.""",
        default=timeout,
    )


def spsdk_apps_common_options(options: FC) -> FC:
    """Common click options.

    Sets --help, --version; provides: `log_level: int` for logging.

    :return: click decorator
    """
    options = click.help_option("--help")(options)
    options = click.version_option(spsdk_version, "--version")(options)
    options = click.option(
        "-vv",
        "--debug",
        "log_level",
        flag_value=logging.DEBUG,
        help="Display more debugging information.",
    )(options)
    options = click.option(
        "-v",
        "--verbose",
        "log_level",
        flag_value=logging.INFO,
        help="Print more detailed information",
    )(options)
    return options


def spsdk_plugin_option(options: FC) -> FC:
    """Plugin click option decorator.

    :return: Click decorator
    """
    return click.option(
        "--plugin",
        required=False,
        help="External python file/package containing a custom plugin implementation.",
    )(options)


def isp_interfaces(
    uart: bool = False,
    usb: bool = False,
    sdio: bool = False,
    lpcusbsio: bool = False,
    buspal: bool = False,
    json_option: bool = True,
    timeout_option: bool = True,
    is_sdp: bool = False,
    use_long_timeout_option: bool = False,
    default_timeout: int = 5000,
) -> Callable:
    """Interfaces Click CLI options.

    :param uart: UART interface, defaults to False
    :param usb: USB interface, defaults to False
    :param sdio: SDIO interface, defaults to False
    :param lpcusbsio: LPCUSBSIO interface, defaults to False
    :param buspal: BUSPAL interface, defaults to False
    :param json_option: add -j option, defaults to True
    :param timeout_option: add timeout option, defaults to True
    :param is_sdp: Specifies whether the ISP interface is meant for SDP(S) protocol
    :param use_long_timeout_option: Use only the long form for timeout (--timeout)
    :param default_timeout: Default timeout to be set when getting the timeout_option
    :return: click decorator
    """

    def decorator(func: Callable[[FC], FC]) -> Callable[[FC], FC]:
        options = []

        if uart:
            options.append(_sdp_uart_option if is_sdp else _uart_option)
        if usb:
            options.append(_usb_option)
        if sdio:
            options.append(_sdio_option)
        if lpcusbsio:
            options.append(_lpcusbsio_option)
        if buspal:
            options.append(_buspal_option)
        if json_option:
            options.append(_json_option)
        if timeout_option:
            option = _timeout_option(use_long_timeout_option, default_timeout)
            options.append(option)
        for option in reversed(options):
            func = option(func)

        return func

    return decorator


class CommandsTreeGroup(click.Group):
    """Custom help formatter, overrides click.Group standard formatter.

    Provides command section in help as command tree

    :param click: click.Group
    """

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Extra format methods for multi methods that adds all the commands after the options.

        :param ctx: click Context
        :param formatter: click HelpFormatter
        """
        root_cmd = _build_command_tree(ctx.find_root().command)
        rows = _get_tree(root_cmd)

        with formatter.section(gettext("Commands")):
            formatter.width = 160
            formatter.write_dl(rows, col_max=80)


class GroupAliasedGetCfgTemplate(click.Group):
    """Alias for get-cfg-template click group extension.

    Temporary class to handle deprecated 'get-cfg-template' command to provide
    better user experience.
    """

    # pylint: disable=inconsistent-return-statements    # ctx.fail at the end terminates the command call
    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        """Override original click get_command function to implement alias for get-cfg-template obsolete command.

        :param ctx: click Context
        :param cmd_name: Requested command name
        :return: Suitable command representation.
        """
        command = click.Group.get_command(self, ctx, cmd_name)
        if command is not None:
            return command
        if cmd_name == "get-cfg-template":
            click.secho(
                "The 'get-cfg-template' is deprecated command, use 'get-template' instead of.",
                fg="yellow",
                bold=True,
                err=True,
            )
            return click.Group.get_command(self, ctx, "get-template")

        ctx.fail(f"Not supported command: '{cmd_name}'")


class CommandsTreeGroupAliasedGetCfgTemplate(CommandsTreeGroup, GroupAliasedGetCfgTemplate):
    """Mix of Command Tree and get-cfg-template alias."""


def _get_tree(
    command: _CommandWrapper,
    rows: Optional[List] = None,
    depth: int = 0,
    is_last_item: bool = False,
    is_last_parent: bool = False,
    parent_prefix: str = "",
) -> Sequence[Tuple[str, str]]:
    """Generate tree of commands to be used with Click HelpFormatter.

    :param command: command wrapper
    :param rows: list of str lines to be printed, defaults to None
    :param depth: tree depth, defaults to 0
    :param is_last_item: last item has different formatting, defaults to False
    :param is_last_parent: last parent item, defaults to False
    :param parent_prefix: visual prefix used by parent node
    :return: definition list to be used with click HelpFormatter
    """
    if rows is None:
        rows = []
    if depth == 0:
        prefix = ""
        tree_item = ""
    else:
        prefix = "    " if is_last_parent else "│   "
        tree_item = "└── " if is_last_item else "├── "

    parent_prefix = parent_prefix + (prefix if depth > 1 else "")
    col1 = parent_prefix + tree_item + command.name
    col2 = str()
    doc: str = command.command.__doc__
    if doc:
        formatted_doc = doc.partition("\n")[0]  # take just first line of doc
        # truncate length to be compliant with max width
        formatted_doc = formatted_doc[:78] + (formatted_doc[78:] and "..")
        col2 += formatted_doc

    rows.append((col1, col2))

    for i, child in enumerate(sorted(command.children, key=lambda x: x.name)):
        _get_tree(
            child,
            rows,
            depth=(depth + 1),
            is_last_item=(i == (len(command.children) - 1)),
            is_last_parent=is_last_item,
            parent_prefix=parent_prefix,
        )
    return rows
